function create( modelName, creationMode )
% autosar.api.create instantiate the mapping and AUTOSAR Properties
% information for a Simulink model
% 
% autosar.api.create(MDLNAME, MODE) instantiates the Simulink mapping and
% AUTOSAR Properties for Simulink model MDLNAME. MODE is an optional
% parameter of either 'init' (default) or 'default'. When MODE is set to
% 'init' an AUTOSAR Software component is created and is mapped to
% the Simulink model. When MODE is set to 'default' the Simulink inports 
% and outports are additionally mapped with defaults created for the 
% corresponding AUTOSAR Properties.
%
% Examples:
%
% To create the default AUTOSAR Properties and mapping for a model
%    autosar.api.create('rtwdemo_autosar_counter', 'default');
%
% See also autosar.api.getAUTOSARProperties autosar.api.getSimulinkMapping



%   Copyright 2012-2014 The MathWorks, Inc.
    
    if nargin < 2
        % Default creationMode
        creationMode = 'init';
    end

    % make sure that model is loaded in memory
    systems = find_system('type', 'block_diagram', 'name', modelName);
    if isempty(systems)
        DAStudio.error('RTW:autosar:mdlNotLoaded', modelName);
    end
    
    % Is it an appropriate target?
    isCompliant = strcmp(get_param(modelName,'AutosarCompliant'), 'on');
    if ~isCompliant
        DAStudio.error('RTW:autosar:nonAutosarCompliant');
    end
    
    % Check modeling style is valid
    autosar.validation.ClientServerValidator.checkValidRunnableConfig(modelName);
    
    % Close the UI as creating a new metamodel will make the UI stale.
    autosar_ui_close(modelName);
    
    switch creationMode
      case 'init'
        % Create initial valid mapping and meta-model

        mmgr = get_param(modelName, 'MappingManager');
        mapping = mmgr.getActiveMappingFor('AutosarTarget');

        if isa(mapping, 'Simulink.AutosarTarget.ModelMapping')
            % As mapping already exists we reset
            mapping.unmap();
            mapping.AUTOSAR_ROOT = [];
        end

        i_createEmptyAutosarMapping(modelName);

        % Add Default AtomicComponent
        componentBuilder = autosar.ui.wizard.builder.Component(modelName, modelName);
        compQName = [componentBuilder.ComponentPackage, '/', componentBuilder.ComponentName];

        dataObj = autosar.api.getAUTOSARProperties(modelName, true);
        dataObj.addComponent(compQName);
      
      case 'default'
        % Setup stage name
        Simulink.output.Stage(...
            message('autosarstandard:ui:uiWizardNoMetaModelOption1').getString(), ...
            'ModelName', modelName, 'UIMode', false);
        
        % Override previous mapping and meta-model with default values
        
        i_createEmptyAutosarMapping(modelName);        
        
        % set defaults using wizard functionality
        componentBuilder = autosar.ui.wizard.builder.Component(modelName, modelName);

        modelH = get_param(modelName, 'Handle');
        componentBuilder.setDefaultConfiguration(modelH);
        
        % configure m3iModel
        m3iModel = autosar.api.Utils.m3iModel(modelName);
        t = M3I.Transaction(m3iModel);
        componentBuilder.populateMetamodel(m3iModel, modelName);
        t.commit();
        
        % configure mapping
        componentBuilder.buildMapping(modelName, 'cmdline');
        
      otherwise
        DAStudio.error('RTW:autosar:invalidCreationMode', creationMode);
    end
    
    % Dirty the model
    set_param(modelName,'Dirty','on');


function i_createEmptyAutosarMapping(modelName)
    % Create empty AutosarTarget mapping. 

    mmgr = get_param(modelName, 'MappingManager');
    [~] = Simulink.AutosarTarget.Component('', ''); % workaround g866593
    mapping = mmgr.getActiveMappingFor('AutosarTarget');
    if ~isempty(mapping)
        mapping.unmap();
        mmgr.deleteMapping(mapping);
    end
    mmgr.createMapping(modelName, 'AutosarTarget');
    mmgr.activateMapping(modelName);
    mapping = mmgr.getActiveMappingFor('AutosarTarget');
    
    m3iModel = Simulink.metamodel.foundation.Factory.createNewModel();
    
    % Start m3i processing
    t = M3I.Transaction(m3iModel);
    m3iModel.Name = 'AUTOSAR';
    % Create the AUTOSAR package
    autosarPkg = Simulink.metamodel.arplatform.common.AUTOSAR(m3iModel);
    autosarPkg.Name = 'AUTOSAR';
    m3iModel.RootPackage.append(autosarPkg);    
    % End m3i processing
    t.commit();      
    
    mapping.AUTOSAR_ROOT = m3iModel;

    % remove AutosarInterface from model
    set_param(modelName, 'RTWFcnClass', []);
    
    
    % Set PreSave Fcn
    b = get_param(modelName,'Object');
    if ~b.hasCallback('PreSave', 'AutosarSaveAsMdl')
        b.addCallback('PreSave','AutosarSaveAsMdl', @()autosarcore.checkSaveAsMdl(b.Handle));
    end

% LocalWords:  MDLNAME
