classdef BasicSoftwareCaller < handle
% Helper class for function-caller mask to use to auto-create, auto-map function-caller properties

%   Copyright 2016 The MathWorks, Inc.

    methods(Static, Access=public)

        function updateBlock(blkPath)
            % Called when setting mask parameters
            autosar.bsw.BasicSoftwareCaller.updateFunctionCaller(blkPath);
        end
        
        function operationCallback(blkPath)
            serviceImpl = autosar.bsw.BasicSoftwareCaller.getServiceImpl(blkPath);
            serviceImpl.operationCallback(blkPath);
        end
        
        function dataTypeCallback(blkPath)
            serviceImpl = autosar.bsw.BasicSoftwareCaller.getServiceImpl(blkPath);
            serviceImpl.dataTypeCallback(blkPath);
        end
        
        function serviceBlocks = find(sys)
            % Find basic software blocks
            serviceBlocks = find_system( sys, ...
                                         'RegExp', 'on', ...
                                         'FollowLinks', 'on', ...
                                         'LookUnderMasks', 'on', ...
                                         'BlockType', 'FunctionCaller', ...
                                         'ServiceImpl', '.*');
        end
        
        function isbswBlock = isBSWCallerBlock(block)
            isbswBlock = ~isempty(autosar.bsw.BasicSoftwareCaller.find(block));
        end
        
        function syncModel( mdlName )
            % Synchronize AUTOSAR Properties and Simulink-AUTOSAR mapping of a model with
            % basic software blocks.

            serviceBlocks = autosar.bsw.BasicSoftwareCaller.find(mdlName);
            for ii=1:length(serviceBlocks)
                autosar.bsw.BasicSoftwareCaller.syncBlock(serviceBlocks{ii});
            end
        end
        
        function blkPath = createBlock(sys, serviceImplStr)
            % Create BasicSoftwareCaller block
            
            serviceImpl = eval(serviceImplStr);
            
            blkName = serviceImpl.getType();
            blkPath = [sys, '/', blkName];
            blkPath = add_block('built-in/FunctionCaller', blkPath, 'MakeNameUnique', 'on');
            
            % Mask block
            mo = get_param(blkPath, 'MaskObject');
            if isempty(mo)
                mo = Simulink.Mask.create(blkPath);
            end

            % Mask help
            mo.Description    = serviceImpl.getDescription();
            mo.Help           = serviceImpl.getHelp();
            mo.Type           = serviceImpl.getType();

            % Mask initialization
            mo.Initialization = 'autosar.bsw.BasicSoftwareCaller.updateBlock(gcb);';
            
            % Mask needs to be self modifiable to modifying underlying
            % Function Caller
            mo.SelfModifiable = 'on';
            
            % Remove and then add mask parameters
            mo.removeAllParameters();
            operations = serviceImpl.getOperations();
            defaultOperation = operations{1};
            if serviceImpl.hasDatatypeParameter() || serviceImpl.hasArgumentSpecificationParameter()
                operationCallbackStr = 'autosar.bsw.BasicSoftwareCaller.operationCallback(gcb);';
            else
                operationCallbackStr = '';
            end
            mo.addParameter('Type', 'edit', ...
                            'Name', 'PortName',...
                            'Prompt', 'Client port name:', ...
                            'Value', serviceImpl.getDefaultClientPortName',...
                            'Tunable','on',...
                            'Evaluate','off',...
                            'Hidden','off');
            mo.addParameter('Type', 'popup', ...
                            'TypeOptions', operations, ...
                            'Name', 'Operation',...
                            'Prompt', 'Operation:', ...
                            'Tunable','on',...
                            'Evaluate','off',...
                            'Callback', operationCallbackStr,...
                            'Hidden','off');
            if serviceImpl.hasDatatypeParameter()
                mo.addParameter('Type', 'edit',...
                                'Name', 'Datatype',...
                                'Value', serviceImpl.getDefaultDatatype(defaultOperation)',...
                                'Tunable','on',...
                                'Evaluate','off',...
                                'Callback', 'autosar.bsw.BasicSoftwareCaller.dataTypeCallback(gcb);',...
                                'Visible', serviceImpl.getDatatypeVisibility(defaultOperation),...
                                'Hidden','off');
            end
            if serviceImpl.hasArgumentSpecificationParameter()
                mo.addParameter('Type', 'edit',...
                                'Name', 'ArgumentSpecification',...
                                'Prompt', 'Argument specification (e.g. int8(1)):',...
                                'Value', serviceImpl.getDefaultArgumentSpecification(defaultOperation)',...
                                'Tunable','off',...
                                'Evaluate','off',...
                                'Visible', serviceImpl.getArgumentSpecificationVisibility(defaultOperation),...
                                'Hidden','off');
            end
            mo.addParameter('Type', 'edit', ...
                            'Name', 'st',...
                            'Prompt', 'Sample time (-1 for inherited):', ...
                            'Value', '-1',...
                            'Tunable','on',...
                            'Evaluate','on',...
                            'Hidden','off');
            mo.addParameter('Type', 'edit', ...
                            'Name', 'ServiceImpl',...
                            'Prompt', 'Service Implementation', ...
                            'Value', serviceImplStr,...
                            'Tunable','on',...
                            'Evaluate','on',...
                            'Visible','off',...
                            'Hidden','off');
            
            % Display just "type" rather than "type (mask)(link)"
            dialogControls = mo.getDialogControls();
            for ii=1:length(dialogControls)
                dc = dialogControls(ii);
                if strcmp(dc.Name, 'DescGroupVar')
                    dc.Prompt = serviceImpl.getType();
                    break
                end
            end
                        
           % Set underlying function-caller parameters
           set_param(blkPath, 'SampleTime', 'st');
           
           
           % Resize block
           width = 325;
           height = 80;
           pos = get_param(blkPath, 'Position');
           pos(3) = pos(1) + width;
           pos(4) = pos(2) + height;
           set_param(blkPath, 'Position', pos);
           
           % Run the initialization commands on the block so block
           % parameters are fully initialized
           Simulink.Block.eval(get_param(blkPath, 'Handle'));
        end

        function val = boolean_spec(val)
        % Function to define boolean (user may have shadowed MATLAB function).
            val = boolean(val);
        end
        
        function val = uint8_spec(val)
        % Function to define uint8 (user may have shadowed MATLAB function).
            val = uint8(val);
        end
       
        function val = uint16_spec(val)
        % Function to define uint16 (user may have shadowed MATLAB function).
            val = uint16(val);
        end
        
        function val = uint32_spec(val)
        % Function to define uint32 (user may have shadowed MATLAB function).
            val = uint32(val);
        end
        
    end
    
    methods(Static, Access=private)
        function serviceImpl = getServiceImpl(blkPath)
            serviceImpl = eval(get_param(blkPath, 'ServiceImpl'));
        end

        function syncBlock(blkPath)
            % Called when syncing block with model
            autosar.bsw.BasicSoftwareCaller.updateAUTOSARProperties(blkPath);
            autosar.bsw.BasicSoftwareCaller.updateSimulinkMapping(blkPath);
        end

        
        function updateFunctionCaller(blkPath)
            % Get Mask parameters
            portName = get_param(blkPath, 'PortName');
            operationName = get_param(blkPath, 'Operation');
            
            % check portName is a valid identifier. Note that operationName
            % is chosen from a drop-down menu and thus no need to validate
            autosar.api.Utils.checkQualifiedName(bdroot(blkPath),...
                portName, 'shortname');
            
            % get the implementation
            serviceImpl = autosar.bsw.BasicSoftwareCaller.getServiceImpl(blkPath);
            
            % run operation and dataType callbacks. This is necessary
            % because changing the mask parameters using set_param does not
            % call the parameter callbacks, so we have to run them from the
            % mask initialization callback
            serviceImpl.operationCallback(blkPath);
            serviceImpl.dataTypeCallback(blkPath);
            
            % Set Function Caller
            serviceImpl.updateFunctionCaller(blkPath, portName, operationName);
        end
        
        function updateAUTOSARProperties(blkPath)
            % Update AUTOSAR Properties
            
            % Copy interface from service library
            autosar.bsw.BasicSoftwareCaller.deepCopyInterface(blkPath);

            % Update port name
            autosar.bsw.BasicSoftwareCaller.updatePortName(blkPath);
        end
        
        
        function updateSimulinkMapping(blkPath)
        % Map function caller

            mdlName = bdroot(blkPath);
            assert(autosar.api.Utils.isMapped(mdlName), 'Expected model to be mapped to AUTOSAR');

            portName = get_param(blkPath, 'PortName');
            operationName = get_param(blkPath, 'Operation');
            
            % Sync function caller
            mapping = autosar.api.Utils.modelMapping(mdlName);
            mapping.sync();

            % Map function caller
            fcnCaller = [portName, '_', operationName];
            slMapping = autosar.api.getSimulinkMapping(mdlName);
            slMapping.mapFunctionCaller(fcnCaller, portName, operationName);
        end

        
        function deepCopyInterface(blkPath)
            % Update interface
            
            serviceImpl = autosar.bsw.BasicSoftwareCaller.getServiceImpl(blkPath);

            mdlName = bdroot(blkPath);
            dataObj = autosar.api.getAUTOSARProperties(mdlName, true);
            interfaceName = serviceImpl.getInterfaceName();
            interfaces = dataObj.find([], 'ClientServerInterface', 'Name', interfaceName, 'PathType', 'FullyQualified');
            if isempty(interfaces)
                serviceImpl.deepCopyInterface(mdlName, [])
            end

        end

        function updatePortName(blkPath)
            % Update port name
           
            serviceImpl = autosar.bsw.BasicSoftwareCaller.getServiceImpl(blkPath);
            
            % Get Mask parameters
            portName = get_param(blkPath, 'PortName');

            mdlName = bdroot(blkPath);
            dataObj = autosar.api.getAUTOSARProperties(mdlName, true);
            component = dataObj.get('XmlOptions', 'ComponentQualifiedName');
            ports = dataObj.find(component, 'Port', 'Name', portName);
            if isempty(ports)
                interfaceName = serviceImpl.getInterfaceName();
                interfaces = dataObj.find([], 'ClientServerInterface', 'Name', interfaceName, 'PathType', 'FullyQualified');
                dataObj.add(component, 'ClientPorts', portName, 'Interface', interfaces{1});
            end
        end

    end
    
end

% LocalWords:  bsw shortname autosarspkglib
