%   Copyright 2014-2016 The MathWorks, Inc.
classdef ClientServerValidator < autosar.validation.PhasedValidator
% Graphical checks for AUTOSAR Client-Server modeling. 

    properties(Constant, Access=private)
        slFcnBlkTypeName = 'Simulink function block';
        fcnCallerBlkTypeName = 'function caller block';
        ValidNvMServiceNames = {'ReadBlock', 'WriteBlock', 'RestoreBlockDefaults'};
    end
    
    methods (Access=protected)

        function verifyInitial(this, hModel)
            % Check simulink function block mappings
            msg = this.verifySimulinkFunctionBlockMapping(hModel);
            
            % Check caller block mappings.
            tmpMsg = this.verifyFunctionCallerBlockMapping(hModel);
            for i = 1:length(tmpMsg.cause)
                msg = msg.addCause(tmpMsg.cause{i});
            end
            
            % Check if simulink function and caller block are in the same
            % model.
            tmpMsg = this.verifyNoCallerAndFcnInSameModel(hModel);
            for i = 1:length(tmpMsg.cause)
                msg = msg.addCause(tmpMsg.cause{i});
            end
            
            % Throw errors
            if ~isempty(msg.cause)
                throw(msg);
            end
        end
        
        function verifyPostProp(this, hModel)
            this.verifyClientArg(hModel);
            msg = this.verifyGlobalSimulinkFunctionCalls(hModel);
            if ~isempty(msg.cause)
                throw(msg);
            end
        end
        
    end
    
    methods (Static, Access=private)
        function msg = verifyNoCallerAndFcnInSameModel(hModel)
            msg = MException('RTW:fcnClass:finish', ...
                DAStudio.message('RTW:fcnClass:finish', 'Multiple causes' ));
            % Skip export function models
            if autosar.validation.ExportFcnValidator.isExportFcn(hModel)
                return;
            end
            modelName = get_param(hModel, 'Name');
            mapping = autosar.api.Utils.modelMapping(modelName);
            fcnNameList = cell(1, length(mapping.ServerFunctions));
            for i = 1:length(mapping.ServerFunctions)
                [~, ~, fcnName] = ...
                    autosar.validation.ClientServerValidator.getBlockInOutParams( ...
                    mapping.ServerFunctions(i).Block);
                fcnNameList{i} = fcnName;
            end
            for i = 1:length(mapping.FunctionCallers)
                [~, ~, fcnName] = ...
                    autosar.validation.ClientServerValidator.getBlockInOutParams( ...
                    mapping.FunctionCallers(i).Block);
                [isDefined, slFcnIdx] = ismember(fcnName, fcnNameList);
                if (isDefined(1))
                    tmpMsg = MException('RTW:autosar:slFcnBlkCallerBlkSameModel', ...
                        DAStudio.message('RTW:autosar:slFcnBlkCallerBlkSameModel', ...
                        mapping.FunctionCallers(i).Block, ...
                        mapping.ServerFunctions(slFcnIdx(1)).Block));
                    msg = msg.addCause(tmpMsg);
                end
            end
        end
        
        function msg = verifyFunctionCallerBlockMapping(hModel)
            modelName = get_param(hModel, 'Name');
            mapping = autosar.api.Utils.modelMapping(modelName);
              
            msg = MException('RTW:fcnClass:finish', ...
                DAStudio.message('RTW:fcnClass:finish', 'Multiple causes' ));
            for i = 1:length(mapping.FunctionCallers)
                if isempty(mapping.FunctionCallers(i).MappedTo.ClientPort) || ...
                        isempty(mapping.FunctionCallers(i).MappedTo.Operation)
                    tmpMsg = MException('RTW:autosar:unmappedClientBlock', ...
                        DAStudio.message('RTW:autosar:unmappedClientBlock', ...
                        mapping.FunctionCallers(i).Block));
                    msg = msg.addCause(tmpMsg);
                    continue;
                end
                m3iOperation = ...
                    autosar.validation.ClientServerValidator.findM3iOpFromPortOpName( ...
                    modelName, mapping.FunctionCallers(i).MappedTo.ClientPort, ...
                    mapping.FunctionCallers(i).MappedTo.Operation);
                % Check if setting on the block is compatible with the
                % operation.
                [isMappable, ~, tmpMsg] = ...
                    autosar.validation.ClientServerValidator.checkFcnCallerMappableToOperation( ...
                    mapping.FunctionCallers(i).Block, m3iOperation);
                if ~isMappable
                    msg = msg.addCause(tmpMsg);
                end
            end
        end
        
        function msg = verifySimulinkFunctionBlockMapping(hModel)
            modelName = get_param(hModel, 'Name');
            mapping = autosar.api.Utils.modelMapping(modelName);
            msg = MException('RTW:fcnClass:finish', ...
                DAStudio.message('RTW:fcnClass:finish', 'Multiple causes' ));
            for i=1:length(mapping.ServerFunctions)
                m3iRunnableObj = autosar.validation.ClientServerValidator.findM3iRunnableFromName( ...
                    modelName, mapping.ServerFunctions(i).MappedTo.Runnable);
                if isempty(m3iRunnableObj)
                    tmpMsg = MException('RTW:autosar:invalidUniqueRunnableName', ...
                        DAStudio.message('RTW:autosar:invalidUniqueRunnableName', ...
                        mapping.ServerFunctions(i).MappedTo.Runnable));
                    msg = msg.addCause(tmpMsg);
                    continue;
                end
                % The simulink function block must match the operation
                % associated with the runnable.
                [isMappable, ~, tmpMsg] = ...
                    autosar.validation.ClientServerValidator.checkSlFcnMappableToRunnable( ...
                    mapping.ServerFunctions(i).Block, m3iRunnableObj);
                if ~isMappable
                    msg = msg.addCause(tmpMsg);
                end
            end
        end

        function verifyClientArg(hModel)
            
            modelName = get_param(hModel, 'Name');
            mapping = autosar.api.Utils.modelMapping(modelName);
            for i = 1:length(mapping.FunctionCallers)
                m3iOperation = ...
                    autosar.validation.ClientServerValidator.findM3iOpFromPortOpName( ...
                    modelName, mapping.FunctionCallers(i).MappedTo.ClientPort, ...
                    mapping.FunctionCallers(i).MappedTo.Operation);
                
                autosar.validation.ClientServerValidator.checkClientArgDataType(...
                    mapping.FunctionCallers(i).Block, m3iOperation);
            end

        end

        function msg = verifyGlobalSimulinkFunctionCalls(hModel)
            % Verify that s-functions (e.g. Stateflow/Embedded MATLAB) don't call global Simulink Functions

            mdlName = get_param(hModel, 'Name');
            msg = MException('RTW:fcnClass:finish', ...
                             DAStudio.message('RTW:fcnClass:finish', 'Multiple causes' ));
            
            % Find s-functions
            rt = sfroot;
            m = rt.find('-isa', 'Simulink.BlockDiagram', '-and', 'Name', mdlName);
            sfcns = m.find('-isa', 'Simulink.SFunction');
            
            for ii=1:length(sfcns)
                h = sfcns(ii).Handle;
                if Simulink.AutosarTarget.ModelMapping.callsGlobalSimulinkFunction(h)
                   blkPath = fileparts(getfullname(h));
                   tmpMsg = MException('autosarstandard:validation:invalidGlobalSimulinkFunctionCall', ...
                                       DAStudio.message('autosarstandard:validation:invalidGlobalSimulinkFunctionCall', ...
                                                        blkPath));
                   msg = msg.addCause(tmpMsg);
               end
            end
        
        end
        
    end
    
    methods (Static, Access=public)

        function checkClientArgDataType(blkPath, m3iOperation)
            
           % Check error status data type
               
           [~, retArgIdx] = autosar.validation.ClientServerValidator.checkArguments(m3iOperation, blkPath);
           if retArgIdx~=-1
               portHandles = get_param(blkPath, 'PortHandles');
               errPortH = portHandles.Outport(retArgIdx);
               autosar.validation.AutosarUtils.checkDataTypeForErrArgPort(errPortH, m3iOperation.containerM3I);
           end
       end

        
        function checkValidRunnableConfig(hModel)
            isExportFcnModel = slprivate('getIsExportFcnModel', hModel);
            slServers =  find_system(hModel, 'FollowLinks', 'on', ...
                                     'blocktype', 'SubSystem', 'IsSimulinkFunction', 'on');
            if ~isempty(slServers) && ~isExportFcnModel
                DAStudio.error('autosarstandard:validation:unsupportedRunnableWithServer');
            end
        end
       
        function [isMappable, errIdx, msg] = checkFcnCallerMappableToOperation(fcnCallerBlock, m3iOperation)
            isMappable = false;
            errIdx = 0;
            msg = [];  %#ok<NASGU>
            [~, outParams, ~] = autosar.validation.ClientServerValidator.getBlockInOutParams(fcnCallerBlock);
            if isempty(m3iOperation) 
                msg = MException('RTW:autosar:fcnCallerOpNameMismatch', ...
                    DAStudio.message('RTW:autosar:fcnCallerOpNameMismatch', ...
                    fcnCallerBlock, ''));
                return;
            else
                m3iInterface = m3iOperation.containerM3I;
                % Check if operation has error argument
                [canHaveError, msg] = autosar.validation.ClientServerValidator.checkOperationHasErrorArg(m3iOperation);
                if ~isempty(msg)
                    return;
                end
                if canHaveError
                    msg = autosar.validation.ClientServerValidator.checkAppErr(m3iInterface);
                    if ~isempty(msg)
                        return;
                    end
                end
                [isCompatible, errIdx, msg] = autosar.validation.ClientServerValidator.checkArguments(m3iOperation, ...
                    fcnCallerBlock);
                if ~isCompatible
                    return;
                else
                    if(slfeature('SlFcnFPCv2_autosar')==0)
                      if length(outParams) > 1 && errIdx >= 1
                        msg = MException('RTW:autosar:errorReturnMustBeOnlyOutput', ...
                          DAStudio.message('RTW:autosar:errorReturnMustBeOnlyOutput', ...
                          fcnCallerBlock));
                        return;
                      end
                    end
                    isMappable = true;
                end
            end
        end
        
        function [isMappable, errIdx, msg] = checkSlFcnMappableToRunnable(slFcnBlock, m3iRunnableObj)
            isMappable = false;
            errIdx = 0;
            msg = [];
            if ~autosar.validation.ExportFcnValidator.isServerSubSys(slFcnBlock)
                return
            end
            
            allM3iOp = autosar.validation.ClientServerValidator.getAllRunnableOperations(m3iRunnableObj);
            if isempty(allM3iOp)
                msg = MException('RTW:autosar:validateIncompatibleOpInvokedEventForRunnable', ...
                    DAStudio.message('RTW:autosar:validateIncompatibleOpInvokedEventForRunnable', ...
                    m3iRunnableObj.Name));
                return;
            end
          
            for i=1:length(allM3iOp)
                m3iOp = allM3iOp{i};
                [canHaveError, msg] = autosar.validation.ClientServerValidator.checkOperationHasErrorArg(m3iOp);
                if ~isempty(msg)
                    return;
                end
                if canHaveError
                    % Check for validity of error codes of interface
                    m3iInterface = m3iOp.containerM3I;
                    msg = autosar.validation.ClientServerValidator.checkAppErr(m3iInterface);
                    if ~isempty(msg)
                        return;    
                    end
                end
                [isCompatible, errIdx, msg] = autosar.validation.ClientServerValidator.checkArguments(m3iOp, ...
                    slFcnBlock);
                if isCompatible
                    [~, outParams, fcnName] = autosar.validation.ClientServerValidator.getBlockInOutParams(slFcnBlock);
                    if strcmp(fcnName, m3iRunnableObj.symbol)
                    if(slfeature('SlFcnFPCv2_autosar')==0)
                      if length(outParams) > 1 && errIdx >= 1
                        msg = MException('RTW:autosar:errorReturnMustBeOnlyOutput', ...
                          DAStudio.message('RTW:autosar:errorReturnMustBeOnlyOutput', ...
                          slFcnBlock));
                        return;
                      end
                    end
                        isMappable = true;
                    else
                        isMappable = false;
                        msg = MException('RTW:autosar:slFcnBlkRunnableSymMismatch', ...
                            DAStudio.message('RTW:autosar:slFcnBlkRunnableSymMismatch', ...
                            slFcnBlock, m3iRunnableObj.Name));
                        break;
                    end
                else
                    isMappable = false;
                    break
                end
            end
        end
        
        function [canHaveError, errmsg] = checkOperationHasErrorArg(m3iOperation)
            % Check if there is an error direction set for operation argument. More
            % than one error direction will be an error condition.
            canHaveError = false;
            errmsg = '';
            errCount = 0;
            
            for index=1:m3iOperation.Arguments.size()
                if m3iOperation.Arguments.at(index).Direction ==  ...
                        Simulink.metamodel.arplatform.interface.ArgumentDataDirectionKind.Error
                    errCount = errCount + 1;
                    if errCount > 1
                        errmsg = MException('autosarstandard:validation:opMultipleReturnArgs', ...
                                    DAStudio.message('autosarstandard:validation:opMultipleReturnArgs',m3iOperation.Name));
                        return;
                    end
                end
            end
            canHaveError = errCount > 0;
        end
        
        function [isCompatible, retArgIdx, errmsg] = checkArguments(m3iOperation, blk)
            
            % Default setting for non-compatible case, should not be
            % altered during the function, unless returning compatible.
            isCompatible = false;
            retArgIdx = -1;
            errmsg = '';
            
            % Always return false if the block type is not function caller
            % or simulink function.
            if ~strcmp(get_param(blk, 'BlockType'), 'FunctionCaller') && ...
                    ~autosar.validation.ExportFcnValidator.isServerSubSys(blk)
                errmsg = MException('RTW:autosar:blkRunnableNotASlFcn', ...
                                    DAStudio.message('RTW:autosar:blkRunnableNotASlFcn',blk,m3iOperation.Name));
                % Not compatible
                return;
            end
            
            [inparam, outparam, ~] = autosar.validation.ClientServerValidator.getBlockInOutParams(blk);
            
            
            % Check for arg name match in block (no duplicates)
            numMatchInParams = 0;
            numMatchOutParams = 0;
            autosarVer = get_param(bdroot(blk), 'AutosarSchemaVersion');
            for i=1:m3iOperation.Arguments.size
                m3iArg = m3iOperation.Arguments.at(i);
               if AR4.isCompatibleSchemaVer(autosarVer)
                   displayFormat = m3iArg.DisplayFormat;
                   [isValid, msg] = autosar.validation.AutosarUtils.checkDisplayFormat(...
                       displayFormat, autosar.api.Utils.getQualifiedName(m3iArg));
                    if ~isValid
                        DAStudio.error('RTW:fcnClass:finish', msg);
                    end
               end
                if(m3iArg.Direction == Simulink.metamodel.arplatform.interface.ArgumentDataDirectionKind.Error)
                    [~,idx] = ismember(m3iArg.Name,outparam);
                    if(idx>0)
                        numMatchOutParams = numMatchOutParams+1;
                        if(slfeature('SlFcnFPCv2_autosar')<=0)
                            % arbitrary ordering is not allowed
                            if(idx~=numMatchOutParams)
                                errmsg = MException('RTW:autosar:blkRunnableArgOrder', ...
                                    DAStudio.message('RTW:autosar:blkRunnableArgOrder', ...
                                    autosar.validation.ClientServerValidator.fcnCallerBlkTypeName, ...
                                    blk, m3iOperation.Name));
                                return; % NOT compatible
                            end
                        end  % SlFcnFPC feat is off
                        retArgIdx = idx;
                        outparam{idx} = '';
                    else
                        errmsg = MException('RTW:autosar:blkRunnableInvalidOutArgName', ...
                          DAStudio.message('RTW:autosar:blkRunnableInvalidOutArgName', ...
                          m3iArg.Name, ...
                          m3iOperation.Name, ...
                          blk));
                        return; % NOT compatible
                    end
                elseif(m3iArg.Direction == Simulink.metamodel.arplatform.interface.ArgumentDataDirectionKind.Out)
                    [~,idx] = ismember(m3iArg.Name,outparam);
                    if(idx>0)
                        outparam{idx} = '';
                        numMatchOutParams = numMatchOutParams+1;
                        if(slfeature('SlFcnFPCv2_autosar')<=0)
                            % arbitrary ordering is not allowed
                            if(idx~=numMatchOutParams)
                                errmsg = MException('RTW:autosar:blkRunnableArgOrder', ...
                                    DAStudio.message('RTW:autosar:blkRunnableArgOrder', ...
                                    autosar.validation.ClientServerValidator.fcnCallerBlkTypeName, ...
                                    blk, m3iOperation.Name));
                                return; % NOT compatible
                            end
                        end  % SlFcnFPC feat is off
                    else
                        errmsg = MException('RTW:autosar:blkRunnableInvalidOutArgName', ...
                          DAStudio.message('RTW:autosar:blkRunnableInvalidOutArgName', ...
                          m3iArg.Name, ...
                          m3iOperation.Name, ...
                          blk));
                        return; % NOT compatible
                    end
                elseif(m3iArg.Direction == Simulink.metamodel.arplatform.interface.ArgumentDataDirectionKind.In)
                    [~,idx] = ismember(m3iArg.Name,inparam);
                    if(idx>0)
                        inparam{idx} = '';
                        numMatchInParams = numMatchInParams+1;
                        if(slfeature('SlFcnFPCv2_autosar')<=0)
                          % arbitrary ordering is not allowed
                          if(idx~=numMatchInParams)
                            errmsg = MException('RTW:autosar:blkRunnableArgOrder', ...
                              DAStudio.message('RTW:autosar:blkRunnableArgOrder', ...
                              autosar.validation.ClientServerValidator.fcnCallerBlkTypeName, ...
                              blk, m3iOperation.Name));
                            return; % NOT compatible
                          end
                        end
                    else
                        errmsg = MException('RTW:autosar:blkRunnableInvalidInArgName', ...
                          DAStudio.message('RTW:autosar:blkRunnableInvalidInArgName', ...
                          m3iArg.Name, ...
                          m3iOperation.Name, ...
                          blk));
                        return; % NOT compatible
                    end
                else
                    assert(m3iArg.Direction == Simulink.metamodel.arplatform.interface.ArgumentDataDirectionKind.InOut);
                    
                    [~,inIdx] = ismember(m3iArg.Name,inparam);
                    [~,outIdx] = ismember(m3iArg.Name,outparam);
                    if(inIdx>0 && outIdx>0)
                        inparam{inIdx} = '';
                        numMatchInParams = numMatchInParams+1;
                        outparam{outIdx} = '';
                        numMatchOutParams = numMatchOutParams+1;
                        if(slfeature('SlFcnFPCv2_autosar')<=0)
                          % arbitrary ordering is not allowed
                          if(inIdx~=numMatchInParams || outIdx~=numMatchOutParams)
                              errmsg = MException('RTW:autosar:blkRunnableArgOrder', ...
                                DAStudio.message('RTW:autosar:blkRunnableArgOrder', ...
                                autosar.validation.ClientServerValidator.fcnCallerBlkTypeName, ...
                                blk, m3iOperation.Name));
                              return; % NOT compatible
                          end
                        end
                    else
                        if(outIdx==0)
                            errmsg = MException('RTW:autosar:blkRunnableInvalidInOutArgName_Output', ...
                              DAStudio.message('RTW:autosar:blkRunnableInvalidInOutArgName_Output', ...
                              m3iArg.Name, ...
                              m3iOperation.Name, ...
                              blk));
                        else
                            errmsg = MException('RTW:autosar:blkRunnableInvalidInOutArgName_Input', ...
                              DAStudio.message('RTW:autosar:blkRunnableInvalidInOutArgName_Input', ...
                              m3iArg.Name, ...
                              m3iOperation.Name, ...
                              blk));
                        end
                        
                        return; % NOT compatible
                    end
                end
            end
            
            
            % Check all inputs were mapped
            if(numMatchInParams~=numel(inparam))
                % Not compatible
                unmappedInputs = ['{''' strjoin(inparam(~cellfun(@isempty,inparam)),''',''') '''}'];
                errmsg = MException('RTW:autosar:blkRunnableMissingInputArgs', ...
                                    DAStudio.message('RTW:autosar:blkRunnableMissingInputArgs',unmappedInputs,blk,m3iOperation.Name));
                return;
            end
            
             % Check return arg
             tmpRetArgIdx = find(~cellfun(@isempty,outparam));
             if(isempty(tmpRetArgIdx))
                 isCompatible = true;
             elseif(numel(tmpRetArgIdx)==1)
                 errmsg = MException('RTW:autosar:blkRunnableAppErrEnumUnspecified', ...
                     DAStudio.message('RTW:autosar:blkRunnableAppErrEnumUnspecified',outparam{tmpRetArgIdx},blk,m3iOperation.Name));
                 % Not compatible
                 return;
             else
                 unmappedOutputs = ['{''' strjoin(outparam(~cellfun(@isempty,outparam)),''',''') '''}'];
                 errmsg = MException('RTW:autosar:blkRunnableMultipleReturnArgs', ...
                     DAStudio.message('RTW:autosar:blkRunnableMultipleReturnArgs',unmappedOutputs,blk,m3iOperation.Name));
                 % Not compatible
                 return;
             end
        end
        
        function m3iOperations = getAllRunnableOperations(m3iRunnableObj)
            m3iOperations = {};
            for i=1:m3iRunnableObj.Events.size
                m3iEvent = m3iRunnableObj.Events.at(i);
                if isa(m3iEvent, ...
                        'Simulink.metamodel.arplatform.behavior.OperationInvokedEvent')
                    if ~isempty(m3iEvent.instanceRef)
                        m3iOp = m3iEvent.instanceRef.Operations;
                        m3iPort = m3iEvent.instanceRef.Port;
                        if isvalid(m3iOp) && isvalid(m3iPort)
                            m3iOperations{end+1} = m3iOp;  %#ok<AGROW>
                        else
                            m3iOperations = []; % If any operation/port combination is not valid, stop
                            break;
                        end
                    end
                end
            end
        end
        
        function [inParams, outParams, fcnName] = getBlockInOutParams(blk)
            % Get the input and output argument names
            inParams = {};
            outParams = {};
            fcnName = '';
            if strcmp(get_param(blk, 'BlockType'), 'FunctionCaller') || ...
                    autosar.validation.ExportFcnValidator.isServerSubSys(blk)
                fcnProto = get_param(blk, 'FunctionPrototype');
                if ~isempty(strfind(fcnProto, '='))
                    outparamStr = regexp(fcnProto, '\[?\s*([^\=\]]+)\]?\s*\=', 'tokens');
                    outparamStr = strtrim(char(outparamStr{1}));
                    if(~isempty(outparamStr))
                      outParams = regexp(outparamStr, '\s*,\s*', 'split');
                    end
                end
                inparamStr = regexp(fcnProto, '\w+\(([^\)]*)\)','tokens');
                inparamStr = strtrim(char(inparamStr{1}));
                if(~isempty(inparamStr))
                  inParams = regexp(inparamStr, '\s*,\s*', 'split');
                end
                fcnName = regexp(fcnProto, '(\w+(\.\w+)?)\s*\(', 'tokens');
                fcnName = char(fcnName{1});
            end
        end
        
        function m3iRunnableObj = findM3iRunnableFromName(modelName, runnableName)
            m3iModel = autosar.api.Utils.m3iModel(modelName);
            m3iRunnableObj = [];
            
            % Get runnable path for ARRunnableName
            dataObj = autosar.api.getAUTOSARProperties(modelName, true);
            componentQualifiedName = dataObj.get('XmlOptions',...
                'ComponentQualifiedName');
            ARRunnablePaths = dataObj.find(componentQualifiedName,...
                'Runnable', 'PathType', 'FullyQualified', ...
                'Name', runnableName);
            
            % Runnable name cannot be duplicated.
            if length(ARRunnablePaths) ~= 1
                return;
            end
            m3iRunnable = autosar.mm.Model.findObjectByNameAndMetaClass( ...
                m3iModel, ARRunnablePaths{1}, ...
                Simulink.metamodel.arplatform.behavior.Runnable.MetaClass());
            if m3iRunnable.size ~= 1
                return;
            end
            m3iRunnableObj = m3iRunnable.at(1);
        end
        
        function m3iOperation = findM3iOpFromPortOpName(modelName, clientPortName, operationName)
            m3iModel = autosar.api.Utils.m3iModel(modelName);
            m3iOperation = [];
            
            % Get operation from client port name and operation name
            dataObj = autosar.api.getAUTOSARProperties(modelName, true);
            componentQualifiedName = dataObj.get('XmlOptions',...
                'ComponentQualifiedName');
            m3iClientPortSeq = autosar.mm.Model.findObjectByNameAndMetaClass( ...
                m3iModel, [componentQualifiedName '/' clientPortName], ...
                Simulink.metamodel.arplatform.port.ClientPort.MetaClass());
            if m3iClientPortSeq.isEmpty()
                return
            elseif m3iClientPortSeq.size() == 1
                for i=1:m3iClientPortSeq.at(1).Interface.Operations.size
                    m3iOp = m3iClientPortSeq.at(1).Interface.Operations.at(i);
                    if strcmp(m3iOp.Name, operationName)
                        m3iOperation = m3iOp;
                        return;
                    end
                end
            else
                DAStudio.error('RTW:autosar:uniqueClientPortNotFound', ...
                        m3iClientPortSeq.size());
            end
        end
        
        function msg = checkAppErr(m3iInterface)
            msg = '';            
            % Validate error codes in interface
            
            % Range check
            for ii=1:m3iInterface.PossibleError.size()
                if m3iInterface.PossibleError.at(ii).errorCode < 0 || ...
                        m3iInterface.PossibleError.at(ii).errorCode > 63
                    % Should never get here. Check is for older models
                    % only
                    msg = MException('autosarstandard:validation:invalidAppErrValue', ...
                        DAStudio.message('autosarstandard:validation:invalidAppErrValue', ...
                                         autosar.api.Utils.getQualifiedName(m3iInterface.PossibleError.at(ii)), ...
                        num2str(m3iInterface.PossibleError.at(ii).errorCode)));
                    return;
                end
            end
        end
        
        function isNvMService = isNvMService(m3iOperation)
        % Does m3iOperation correspond to asynchronous NvM Service
            
            isNvMService = any(strcmp(m3iOperation.Name, ...
                              autosar.validation.ClientServerValidator.ValidNvMServiceNames));

        end
            
    end
end

% LocalWords:  blocktype autosarstandard FP ASl FPC metamodel arplatform Nv
