classdef ExportFcnValidator < autosar.validation.PhasedValidator
% Graphical checks for AUTOSAR export functions. 

%   Copyright 2013-2015 The MathWorks, Inc.
    
    methods (Access=protected)

        function verifyInitial(this, newSubsysHdl, origSSHdl)
            this.verifyIOPortsAreConnected(newSubsysHdl);
            this.verifyGraphicalBlocks(newSubsysHdl, origSSHdl);
            this.verifyIRVs(newSubsysHdl);
            this.verifyFcnCallInportsSampleTime(newSubsysHdl);            
        end

        function verifyPostProp(this, newSubsysHdl, ~)
            % Require compiled model
            this.verifyIRVsTopModel(newSubsysHdl);
            this.verifyGotoFromBlocks(newSubsysHdl);
            this.verifyIRVDataTypes(newSubsysHdl);
            this.verifyConditionallyAssignedAndMergedOutports(newSubsysHdl);
            this.verifyIRVsTopModel(newSubsysHdl);
        end
    end
    
    
    methods (Static, Access=public)
        
        function isExportFcn = isExportFcn(hModel)
            isExportFcn = autosar.validation.ExportFcnValidator.isTopModelExportFcn(hModel) || ...
                          autosar.validation.ExportFcnValidator.isVirtualSubsysExportFcn(hModel);
        end
        
        function isServerSS = isServerSubSys(sys)
            isServerSS = false;
            if (strcmp(get_param(sys, 'BlockType'), 'SubSystem') && ...
                    strcmp(get_param(sys, 'IsSimulinkFunction'), 'on'))
                isServerSS = true;
            end
        end

        function [isAsyncFcnCall, rootFcnCallInportName] = isTopModelAsyncFcnCall(sys)
        % These are the models which are not export functions but have
        % root function-call inport blocks            
            isAsyncFcnCall = false;
            rootFcnCallInportName = [];
            isExportFcn = slprivate('getIsExportFcnModel', bdroot(sys));
            if(~isExportFcn)
                rootFcnCallInport =  find_system(bdroot(sys),  'SearchDepth', 1, ...
                                                 'FollowLinks', 'on', 'blocktype', 'Inport', ...
                                                 'OutputFunctionCall', 'on');
                if ~isempty(rootFcnCallInport)
                    rootFcnCallInportName = rootFcnCallInport{1};
                    isAsyncFcnCall = true;
                end 
            end
        end
        
        function isExportFcn = isTopModelExportFcn(sys)
            isExportFcn = slprivate('getIsExportFcnModel', bdroot(sys));
            if ~isExportFcn
                % Special consideration for autosar
                slServers =  find_system(bdroot(sys),  'SearchDepth', 1, ...
                    'FollowLinks', 'on', 'blocktype', 'SubSystem', ...
                                         'IsSimulinkFunction', 'on');
                if ~isempty(slServers)
                    isExportFcn = true;
                end
            end
        end
        
        function isExportFcn = isVirtualSubsysExportFcn(sys)
        % Classic virtual subsystem export functions
            newSubsysHdl = get_param(bdroot(sys), 'NewSubsystemHdlForRightClickBuild');
            tlcOptions = get_param(bdroot(newSubsysHdl), 'TLCOptions');
            isExportFcn = ~isempty(strfind(tlcOptions, 'aExportFunctionsMode'));
        end
        
        function isModelWideEvent = isModelWideEvent(blkH)
            % Checks if the given block corresponds to a model wide event;
            % e.g., initialization, reset, terminate.

            isModelWideEvent = false;
            if ~strcmp(get_param(blkH, 'BlockType'), 'SubSystem')
                return
            end
            blockList = find_system(blkH, 'SearchDepth', 1, 'LookUnderMasks', 'all', ...
                'FollowLinks', 'on');
            for ii = 1:length(blockList)
                blkH = blockList(ii);
                if strcmp(get_param(blkH, 'BlockType'), 'EventListener') && ...
                   ismember(get_param(blkH, 'EventType'), ...
                            {'Initialize', 'Reset', 'Terminate'})
                    isModelWideEvent = true;
                    return
                end
            end
        end
        
    end
    
    
        
    methods (Static, Access=private)
        
        function verifyIRVs(ssBlkH) 
        % Function: verifyIRVs ================
        % Abstract: Verify IRVs
                        
            isMultiRunnable = autosar.validation.ExportFcnValidator.isMultiRunnable(ssBlkH);
            isExportFcn = autosar.validation.ExportFcnValidator.isExportFcn(ssBlkH);
            %
            % check the exportedFcns conditions for AUTOSAR
            %
            if isMultiRunnable && isExportFcn
                [status, msg] = autosar.validation.p_checkIrvs(ssBlkH, ssBlkH, true);
                if status == 0
                     DAStudio.error('RTW:fcnClass:finish', msg);
                end
            end
        end
        function verifyFcnCallInportsSampleTime(ssBlkH)
            if (autosar.validation.ExportFcnValidator.isTopModelExportFcn(ssBlkH))
                mapping = autosar.api.Utils.modelMapping(bdroot(ssBlkH));
                assert(~isempty(mapping));
                m3iModel = autosar.api.Utils.m3iModel(mapping.Name);
                dataObj = autosar.api.getAUTOSARProperties(mapping.Name, true);
                compQualPath = dataObj.get('XmlOptions', 'ComponentQualifiedName');
                compObjSeq = autosar.mm.Model.findObjectByName(m3iModel, compQualPath);
                assert(compObjSeq.size() > 0);
                compObj = compObjSeq.at(1);
                behaviorObj = compObj.Behavior;
                
                for ii = 1 : numel(mapping.FcnCallInports)
                    blk = mapping.FcnCallInports(ii).Block;
                    runnableName = mapping.FcnCallInports(ii).MappedTo.Runnable;
                    if ~isempty(runnableName)
                        runnableHasTimingEvent = false;
                        for jj = 1 : behaviorObj.Events.size()
                            if isa(behaviorObj.Events.at(jj), 'Simulink.metamodel.arplatform.behavior.TimingEvent') && ...
                                    strcmp(behaviorObj.Events.at(jj).StartOnEvent.Name, runnableName)
                                runnableHasTimingEvent = true;
                                break;
                            end
                        end
                        sample_time_expr = get_param(blk, 'SampleTime');
                        [sampleTime, itExists] = slResolve(sample_time_expr, blk);
                        if runnableHasTimingEvent
                            if itExists 
                                if sampleTime(1) < 0
                                    messageID = 'RTW:autosar:errorInvalidParameter';
                                    DAStudio.error(messageID, runnableName,...
                                        'Sample time', mapping.FcnCallInports(ii).Block,...
                                        'value greater than zero');
                                end
                                if (length(sampleTime) > 1 && sampleTime(2) > 0)
                                    messageID = 'RTW:autosar:errorInvalidParameter';
                                    DAStudio.error(messageID, runnableName,...
                                        'Sample time', mapping.FcnCallInports(ii).Block,...
                                        'zero initial time offset');
                                end
                            end                          
                            
                        else
                            if itExists && sampleTime(1) ~= -1
                                messageID = 'RTW:autosar:errorInvalidParameter';
                                DAStudio.error(messageID, runnableName,...
                                    'Sample time', mapping.FcnCallInports(ii).Block,...
                                    '-1');
                            end
                        end
                    end
                end
            end
        end
        
        function verifyIRVsTopModel(ssBlkH) 
        % Function: verifyIRVs ================
        % Abstract: Verify IRVs
        
            % Check on the lines at this level this was performed in
            % p_checkIrvs 
            lines = find_system(ssBlkH, 'FindAll', 'on', 'SearchDepth', ...
                1, 'FollowLinks', 'on', 'LookUnderMasks', 'all', ...
                'Type', 'line');
            for idx = 1:length(lines)
                lineObj = get_param(lines(idx), 'Object');
                lineName = get_param(lineObj.Handle, 'Name');
                if ~isempty(lineName)
                    if any(lineName >= 128) % Check for ASCII
                        DAStudio.error('autosarstandard:validation:nonAsciiSignal',...
                            lineName);
                    end
                end
                srcBlock = lineObj.SrcBlockHandle;
                dstBlocks = lineObj.DstBlockHandle;
                % Check that line has a source block
                if srcBlock < 0
                    if ~isempty(dstBlocks) && dstBlocks(1) > 0
                        DAStudio.message('RTW:autosar:MultiRunnableLineWithNoSrc', ...
                            get_param(lineObj.DstPortHandle, 'PortNumber'), ...
                            RTW.AutosarInterface.p_getTreatedFullName(dstBlocks(1)));
                    end
                end

                % Check that line has a destination block
                if isempty(dstBlocks) || dstBlocks(1) < 0
                    DAStudio.error('RTW:autosar:MultiRunnableLineWithNoDst', ...
                        get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                        RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                end
            end

            
            if (autosar.validation.ExportFcnValidator.isTopModelExportFcn(ssBlkH))
                mapping = autosar.api.Utils.modelMapping(bdroot(ssBlkH));
                mapping.validateDataTransfers();
            end
        end
        
        
        function verifyGraphicalBlocks(ssBlkH, srcBlkH)
        % Function: checkGraphicalBlocks =============================
        % Abstract: graphical checks

           if strcmp(get_param(srcBlkH, 'Type'), 'block_diagram')
               isLibrary = false;
           else
               isLibrary = strcmp(get_param(srcBlkH, 'LinkStatus'), 'resolved');
           end
           isMultiRunnable = autosar.validation.ExportFcnValidator.isMultiRunnable(ssBlkH);

           % check on the blocks
           if isMultiRunnable
               blks = autosar.validation.ExportFcnValidator.findSystemThruVirtualSubsystems(ssBlkH, 'block');
           else
               blks = ssBlkH;
           end

           hModel = bdroot(ssBlkH);
           for idx = 1:length(blks)
               % From blocks must have matching Goto's
               % For right-click export fcn, the related From must in
               % the same level"
               if strcmp(get_param(blks(idx), 'BlockType'), 'From')
                   blkObj = get_param(blks(idx), 'Object');
                   if isempty(blkObj.GotoBlock.handle)
                       DAStudio.error('RTW:autosar:MultiRunnableLineWithNoGoto', ...
                           RTW.AutosarInterface.p_getTreatedFullName(blks(idx)));
                   end

                   % For right-click export fcn, Goto blocks must have 
                   % local scope and matching froms inside multirunnable
                   % For top level build, we allow global Goto blocks at the
                   % runnable level
               elseif strcmp(get_param(blks(idx), 'BlockType'), 'Goto')
                   if ~autosar.validation.ExportFcnValidator.isTopModelExportFcn(hModel) && ...
                           ~strcmp(get_param(blks(idx), 'TagVisibility'), 'local')
                       DAStudio.error('RTW:autosar:NonLocalGotoInWrapper', ...
                           RTW.AutosarInterface.p_getTreatedFullName(blks(idx)));
                   end
                   blkObj = get_param(blks(idx), 'Object');
                   fromHandles = {blkObj.FromBlocks.handle};
                   if isempty(fromHandles)
                       DAStudio.error('RTW:autosar:MultiRunnableLineWithNoFrom', ...
                           RTW.AutosarInterface.p_getTreatedFullName(blks(idx)));
                   end

               elseif  strcmp(get_param(blks(idx), 'BlockType'), 'SubSystem') && ...
                       strcmp(RTW.AutosarInterface.p_checkTrigSSTypeUsingTrigPortBlock(blks(idx)),'function-call')

                   if ~isLibrary && isMultiRunnable && autosar.validation.ExportFcnValidator.isVirtualSubsysExportFcn(ssBlkH)
                       % only check on MultiRunnable fcn-call subsystems (Children)
                       ssRTWFcnClass = get_param(blks(idx), 'SSRTWFcnClass');
                       if isempty(ssRTWFcnClass)
                           DAStudio.error('RTW:autosar:NoUnconfiguredTopFCSS', ...
                                                  RTW.AutosarInterface.p_getTreatedFullName(blks(idx)));
                       end
                   end

                   % Limiting the following validation checks only to the right-click build case.
                   % TopModel Export functions will not have these validation errors. 
                   if(~autosar.validation.ExportFcnValidator.isTopModelExportFcn(hModel))
                       % Check that function trigger port does not have "States when enabled" set to
                       % "reset" if multirunnable, in order to avoid model_enable functions in the
                       % generated code
                       StatesWhenEnabledResetTriggerPortList = find_system(...
                           blks(idx), 'SearchDepth', 1, 'LookUnderMasks', 'all', ...
                           'FollowLinks', 'on', 'BlockType', 'TriggerPort', ...
                           'TriggerType', 'function-call', 'StatesWhenEnabling', 'reset' );
                       
                       if ~isempty(StatesWhenEnabledResetTriggerPortList) && isMultiRunnable
                           DAStudio.error('RTW:autosar:multirunnableHasStatesWhenEnabledReset', ...
                                          RTW.AutosarInterface.p_getTreatedFullName(...
                                              StatesWhenEnabledResetTriggerPortList(1)));
                       end

                       autosar.validation.ExportFcnValidator.checkForEnabledSubsystemsWithResetConditions(blks(idx), isMultiRunnable);
                   end
                   
                   if ~autosar.validation.ExportFcnValidator.isTopModelExportFcn(hModel)
                       % Avoid inport latches because they end up causing a copy to the input port of
                       % the hidden signal copy blocks
                       inblks = find_system(blks(idx), 'FindAll', 'on', 'SearchDepth', 1, 'BlockType', 'Inport');
                       for idx2 = 1:length(inblks)
                           inBlkH = inblks(idx2);
                           if strcmpi(get_param(inBlkH, 'LatchByDelayingOutsideSignal'), 'on')
                               DAStudio.error('RTW:autosar:multirunnableHasDelayLatch', ...
                                              RTW.AutosarInterface.p_getTreatedFullName(inBlkH));
                           end
                           if strcmpi(get_param(inBlkH, 'LatchInputForFeedbackSignals'), 'on')
                               DAStudio.error('RTW:autosar:multirunnableHasFeedbackLatch', ...
                                              RTW.AutosarInterface.p_getTreatedFullName(inBlkH));
                           end
                       end
                   end
                   
                   % Limiting the following validation checks only to the right-click build case.
                   % TopModel Export functions will not have these validation errors. 
                   if(~autosar.validation.ExportFcnValidator.isTopModelExportFcn(hModel))
                       outblks = find_system(blks(idx), 'FindAll', 'on', 'SearchDepth', 1, 'BlockType', 'Outport');
                       for idx2 = 1:length(outblks)
                           % Check that output port does not have "Output when disabled" set to
                           % "reset", in order to avoid model_disable functions in the
                           % generated code
                           if strcmp( get_param(outblks(idx2), 'OutputWhenDisabled'), 'reset')
                               DAStudio.error('RTW:autosar:multirunnableHasOutputWhenDisabledReset', ...
                                              RTW.AutosarInterface.p_getTreatedFullName(outblks(idx2)));
                           end
                           
                           IC = get_param(outblks(idx2), 'InitialOutput');
                           ICValue = evalinGlobalScope(hModel, IC);
                           typeName = class(ICValue);
                           if isa(ICValue, 'embedded.fi')
                               % dealing with fixed point type
                               if storedInteger(ICValue) ~= 0
                                   DAStudio.error('RTW:autosar:MultiRunnableZeroOutIC', ...
                                                  RTW.AutosarInterface.p_getTreatedFullName(outblks(idx2)));
                               end
                           elseif Simulink.data.isSupportedEnumClass(typeName)
                               % dealing with enum type
                               enumDefaultValue = Simulink.data.getEnumTypeInfo(typeName, 'DefaultValue');
                               
                               if eval(IC) ~= enumDefaultValue
                                   DAStudio.error('RTW:autosar:MultiRunnableZeroOutIC', ...
                                                  RTW.AutosarInterface.p_getTreatedFullName(outblks(idx2)));
                               end
                           else                     
                               if ~strcmp(IC, '[]') && ~strcmp(IC, '0') && ~strcmp(IC, '0.0') && all(double(eval(IC)) ~= 0)
                                   DAStudio.error('RTW:autosar:MultiRunnableZeroOutIC', ...
                                                  RTW.AutosarInterface.p_getTreatedFullName(outblks(idx2)));
                               end
                           end
                           
                           initSrc = get_param(outblks(idx2), 'SourceOfInitialOutputValue');
                           if ~strcmp(initSrc, 'Dialog')
                               DAStudio.error('RTW:autosar:multirunnableHasOutputWithDynamicInit', ...
                                              RTW.AutosarInterface.p_getTreatedFullName(outblks(idx2)));
                           end
                       end
                   end
               end
           end

       end
       
       function verifyIOPortsAreConnected(ssBlkH)
           
           % If multirunnable, check on the ports inside the wrapper; make sure
           % they all have lines connected to them
           isMultiRunnable = autosar.validation.ExportFcnValidator.isMultiRunnable(ssBlkH);
           if isMultiRunnable
               ports = autosar.validation.ExportFcnValidator.findSystemThruVirtualSubsystems(ssBlkH, 'port');
               for idx = 1:length(ports)
                   portObj = get_param(ports(idx), 'Object');
                   parent = portObj.Parent;
                   if get_param(parent, 'Handle') ~= ssBlkH
                       lineH = portObj.Line;
                       isError = false;
                       % If the line does not exist or any of the source or
                       % destination blocks are commented out
                       if lineH < 0
                           if ~(autosar.validation.ExportFcnValidator.isServerSubSys(parent) && ...
                                   strcmp(portObj.PortType, 'trigger'))
                               isError = true;
                           end
                       else
                           allBlkH = cat(1, get_param(lineH,'SrcBlockHandle'), ...
                               get_param(lineH,'DstBlockHandle'));
                           for blkIdx = 1:length(allBlkH)
                               blkH = allBlkH(blkIdx);
                               if (blkH >= 0 && ...
                                   strcmp(get_param(blkH, 'Commented'), 'on'))
                                   isError = true;
                               end
                           end
                       end
                       if isError
                           portType = portObj.PortType;
                           portType(1) = upper(portType(1));
                           DAStudio.error('RTW:autosar:MultiRunnableUnconnectedPort', ...
                               portType, ...
                               portObj.PortNumber,...
                               RTW.AutosarInterface.p_getTreatedFullName(parent));
                       end
                   end
               end
           end

       end
           

       function verifyConditionallyAssignedAndMergedOutports(ssBlkH)
           % Function: checkConditionallyAssignedAndMergedOutports ========================
           % Abstract: [Engine Interface] If initialization mode is classic, check to see if the
           % outports are conditionally assigned and merged.

           isMultiRunnable = autosar.validation.ExportFcnValidator.isMultiRunnable(ssBlkH);
           
           if isMultiRunnable
               runnablesHdls = autosar.validation.ExportFcnValidator.getRunnableHdls(ssBlkH, isMultiRunnable);
               if strcmpi(...
                   get_param(bdroot(ssBlkH), 'UnderspecifiedInitializationDetection'), ...
                   'classic')
                   sess = Simulink.CMI.EIAdapter(1);
                   for idx = 1:length(runnablesHdls)
                       outblks = find_system(runnablesHdls(idx), 'FindAll', 'on', 'SearchDepth', 1, 'BlockType', 'Outport');
                       for idx2 = 1:length(outblks)
                           outportNum = str2double(get_param(outblks(idx2), 'Port'));
                           ssPorts = get_param(runnablesHdls(idx),'PortHandles');
                           ssOutportH = ssPorts.Outport(outportNum);
                           ssLineH = get_param(ssOutportH, 'Line');
                           ssLineObj = get_param(ssLineH, 'Object');
                           dsts = autosar.validation.walkThruVirtualBlockToDsts(...
                               ssLineObj, ssBlkH, false);
                           mergeBlk = [];
                           for dstIdx = 1:length(dsts)
                               if strcmpi(get_param(dsts(dstIdx), 'BlockType'), 'Merge')
                                   mergeBlk = dsts(dstIdx);
                                   break;
                               end
                           end
                           if ~isempty(mergeBlk)
                               
                               uddobj = get_param(outblks(idx2), 'UDDObject');
                               srcIsConditional = uddobj.srcExecutesConditionally;
                               if srcIsConditional == 1 % possibly
                                   DAStudio.warning(...
                                       'RTW:autosar:MergedIRVConditionalWrite', ...
                                       outportNum, ...
                                       RTW.AutosarInterface.p_getTreatedFullName(get_param(outblks(idx2), 'Parent')), ...
                                       RTW.AutosarInterface.p_getTreatedFullName(mergeBlk));
                               elseif srcIsConditional == 2 % definitely
                                   DAStudio.error(...
                                       'RTW:autosar:MergedIRVConditionalWrite', ...
                                       outportNum, ...
                                       RTW.AutosarInterface.p_getTreatedFullName(get_param(outblks(idx2), 'Parent')), ...
                                       RTW.AutosarInterface.p_getTreatedFullName(mergeBlk));
                               end
                           end
                       end
                   end
                   delete(sess);
               end

           end
       end
           
       function verifyGotoFromBlocks(ssBlkH)
               
            isMultiRunnable = autosar.validation.ExportFcnValidator.isMultiRunnable(ssBlkH);
            if isMultiRunnable
               runnablesHdls = autosar.validation.ExportFcnValidator.getRunnableHdls(ssBlkH, isMultiRunnable);
                
               fromBlocks = find_system(...
                   ssBlkH, 'LookUnderMasks', 'all', 'FollowLinks', 'on', 'BlockType', 'From');
               for idx = 1:length(fromBlocks)
                   fromBlock = fromBlocks(idx);
                   gotoBlock = get_param(fromBlock, 'GotoBlock');
                   gotoBlock = gotoBlock.handle;
                   if ~isempty(gotoBlock)
                       fromRunnable = autosar.validation.ExportFcnValidator.findRunnableOwner(fromBlock, runnablesHdls);
                       gotoRunnable = autosar.validation.ExportFcnValidator.findRunnableOwner(gotoBlock, runnablesHdls);
                       if ~isequal(fromRunnable, gotoRunnable)
                           DAStudio.error(...
                               'RTW:autosar:InvInterSysConnAutosarRunnable', ...
                               RTW.AutosarInterface.p_getTreatedFullName(gotoRunnable), ...
                               RTW.AutosarInterface.p_getTreatedFullName(gotoBlock), ...
                               RTW.AutosarInterface.p_getTreatedFullName(fromBlock));
                       end
                   end
               end
               
               if autosar.validation.ExportFcnValidator.isTopModelExportFcn(ssBlkH)
                   return
               end
               
               fromBlocks = find_system(...
                   get_param(get_param(ssBlkH, 'Parent'), 'Handle'), ...
                   'SearchDepth', 1, 'FollowLinks', 'on', 'BlockType', 'From');
               for idx = 1:length(fromBlocks)
                   fromBlock = fromBlocks(idx);
                   gotoBlock = get_param(fromBlock, 'GotoBlock');
                   gotoBlock = gotoBlock.handle;
                   fromRunnable = autosar.validation.ExportFcnValidator.findRunnableOwner(fromBlock, runnablesHdls);
                   assert(isempty(fromRunnable), 'owner should be tempModel');
                   gotoRunnable = autosar.validation.ExportFcnValidator.findRunnableOwner(gotoBlock, runnablesHdls);
                   % If the goto is outside the multirunnable canvas, OK: both
                   % fromRunnable and gotoRunnable will be empty
                   if ~isequal(fromRunnable, gotoRunnable)
                       DAStudio.error(...
                           'RTW:autosar:InvInterSysConnAutosarRunnable2', ...
                           RTW.AutosarInterface.p_getTreatedFullName(gotoRunnable), ...
                           RTW.AutosarInterface.p_getTreatedFullName(gotoBlock));
                   end
               end
            end
           
           
       end
       
       function verifyIRVDataTypes(ssBlkH)
           % Function: verifyIrvDataTypes ========================
           % Abstract: [Compiled model] 

           hModel = bdroot(ssBlkH);
           cs = autosar.validation.ConfigSetValidator.getActiveConfigSet(hModel);
           maxShortNameLength = get_param(cs, 'AutosarMaxShortNameLength');
            
           isSupportMatrixIOAsArray = strcmpi(...
               get_param(cs,'AutosarMatrixIOAsArray'), 'on');
           
           isMultiRunnable = autosar.validation.ExportFcnValidator.isMultiRunnable(ssBlkH);

           if autosar.validation.ExportFcnValidator.isVirtualSubsysExportFcn(ssBlkH)
              fromBlocks = find_system(...
                  get_param(get_param(ssBlkH, 'Parent'), 'Handle'), ...
                  'SearchDepth', 1, 'FollowLinks', 'on', 'BlockType', 'From');
                  assert(isempty(fromBlocks), 'should not be any From blocks because this can lead to a synthesized outport, which is not covered by the AUTOSAR configuration');
           end
                  
           if isMultiRunnable
               lines = autosar.validation.ExportFcnValidator.findSystemThruVirtualSubsystems(ssBlkH, 'line');
               for idx = 1:length(lines)
                   lineObj = get_param(lines(idx), 'Object');
                   srcBlock = lineObj.SrcBlockHandle;
                   dstBlocks = lineObj.DstBlockHandle;

                   % Interrunnable variables cannot be complex. It also cannot be arrays,
                   % or bus if the schema version is before 4.0.
                   if strcmp(RTW.AutosarInterface.p_checkTrigSSTypeUsingTrigPortBlock(srcBlock), 'function-call') || ...
                           strcmp(get_param(srcBlock, 'BlockType'), 'Merge')
                       for dstIdx = 1:length(dstBlocks)
                           if strcmp(RTW.AutosarInterface.p_checkTrigSSTypeUsingTrigPortBlock(dstBlocks(dstIdx)),...
                                     'function-call')
                               % check that it is a scalar noncomplex
                               if lineObj.getSourcePort.CompiledPortComplexSignal
                                   DAStudio.error('RTW:autosar:MultiRunnableInterrunnableComplex', ...
                                                          get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                                                          RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                               end
                               
                               % Only checks if IRV is array or struct if it is pre
                               % AUTOSAR 4.0
                               if ~AR4.isCompatibleSchemaVer( autosar.SchemaVersionHandler.instance().getSchemaVersion() )
                                   if lineObj.getSourcePort.CompiledPortWidth > 1
                                       DAStudio.error('RTW:autosar:MultiRunnableInterrunnableArray', ...
                                                              get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                                                              RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                                      
                                   end
                                   
                                   if existsInGlobalScope(hModel, lineObj.getSourcePort.CompiledPortDataType) && ...
                                           evalinGlobalScope(hModel,['isa(' lineObj.getSourcePort.CompiledPortDataType ',''Simulink.Bus'')'])
                                       DAStudio.error('RTW:autosar:MultiRunnableInterrunnableBus', ...
                                                              get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                                                              RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                                      
                                   end
                               else
                                   % For AUTOSAR 4.0 check if the datatype is valid.
                                   autosar.validation.AutosarUtils.checkDataType(hModel, lineObj.getFullName, ...
                                                   get_param(lineObj.SrcPortHandle, ...
                                                             'CompiledPortDataType'), ...
                                                   maxShortNameLength, isSupportMatrixIOAsArray);
                                   
                               end

                               if ~strcmp(lineObj.getSourcePort.CompiledRTWStorageClass, 'Auto')
                                   DAStudio.error('RTW:autosar:MultiRunnableInterrunnableNonAutoStorageClass', ...
                                                          lineObj.Name, ...
                                                          get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                                                          RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                                   
                               end
                           end
                       end
                   end
               end
           end
       end   

    end
    
    methods (Static, Access=public)
       
       % Function: convertErrMsgToUseActModel ==================================
       % Abstract:
       %   RightClickBuild creates a hidden model. We need to transform the
       %   error message with the hidden model name to an error message with the
       %   actual model name. When runValidation is called from the command line,
       %   cache will contain the original model name. Otherwise this functionality
       %   needs to be called higher up.
       function msg = convertErrMsgToUseActModel( hSrc, msg )
           if ~isempty(hSrc.cache) && hSrc.SubsysBlockHdl ~= -1
               hiddenModelName = get_param(hSrc.cache.ModelHandle, 'Name');
               actualModelName = get_param(hSrc.SubsysBlockHdl, 'Parent');
               msg = regexprep(msg, ['''', hiddenModelName, '/'], ...
                               ['''',actualModelName,'/']);
           end
           msg = strrep(msg, '\n', ' ');
       end
    
    end
    
    methods (Static, Access=private)

       function objs = findSystemThruVirtualSubsystems(ctxt, searchType)
           objs = [];
           blockList = find_system(ctxt, 'SearchDepth',...
                                   1, 'LookUnderMasks', 'all', 'FollowLinks', 'on');
           for i = 1:length(blockList)
               ssBlkH = blockList(i);
               if ~isequal(ssBlkH, ctxt) && strcmp(get_param(ssBlkH, 'BlockType'),'SubSystem')
                   if strcmpi(get_param(ssBlkH, 'Virtual'), 'on')
                       ssObjs = autosar.validation.ExportFcnValidator.findSystemThruVirtualSubsystems(ssBlkH, searchType);
                       objs = [objs; ssObjs]; %#ok
                   end
               end
           end
           localObjs = find_system(ctxt, ...
                                   'FindAll', 'on', ...
                                   'SearchDepth', 1, ...
                                   'LookUnderMasks', 'all', ...
                                   'FollowLinks', 'on', ...
                                   'Type', searchType);
           objs = [objs; localObjs];
       end


       function checkForEnabledSubsystemsWithResetConditions( sys, isMultiRunnable )

       % only need to check for multi-runnables
           if ~isMultiRunnable
               return
           end
           
           % Find all subsystems
           subsystemBlks = find_system(sys, 'LookUnderMasks', 'all', 'FollowLinks', 'on', 'BlockType', 'SubSystem');

           for idx = 1:length(subsystemBlks)
               % Check for enable port
               enablePortBlks = find_system(subsystemBlks(idx), 'SearchDepth', 1, 'LookUnderMasks', 'all', ...
                                            'FollowLinks', 'on', 'BlockType', 'EnablePort');
               
               if isempty(enablePortBlks)
                   % subsystem doesn't have enable port so skip checks
                   continue
               end

               % Check that enable port does not have "States when enabled" set to
               % "reset" if multirunnable, in order to avoid model_enable functions in the
               % generated code
               StatesWhenEnabledResetPortList = find_system(subsystemBlks(idx), 'SearchDepth', 1, 'LookUnderMasks', 'all', ...
                                                            'FollowLinks', 'on', 'BlockType', 'EnablePort', ...
                                                            'StatesWhenEnabling', 'reset' );
               
               if ~isempty(StatesWhenEnabledResetPortList)
                   DAStudio.error('RTW:autosar:multirunnableHasStatesWhenEnabledReset', ...
                                  RTW.AutosarInterface.p_getTreatedFullName(StatesWhenEnabledResetPortList(1)));
               end                             
               
               outblks = find_system(subsystemBlks(idx), 'FindAll', 'on', 'SearchDepth', 1, 'BlockType', 'Outport');
               for idx2 = 1:length(outblks)
                   % Check that output port does not have "Output when disabled" set to
                   % "reset" if multirunnable, in order to avoid model_disable functions in the
                   % generated code
                   
                   if strcmp( get_param(outblks(idx2), 'OutputWhenDisabled'), 'reset')
                       DAStudio.error('RTW:autosar:multirunnableHasOutputWhenDisabledReset', ...
                                      RTW.AutosarInterface.p_getTreatedFullName(outblks(idx2)));
                   end
               end
           end
       end

       function runnable = findRunnableOwner(block, runnablesHdls)
           parent = get_param(get_param(block, 'Parent'), 'Handle');
           parentType = get_param(parent, 'Type');
           if ~strcmpi(parentType, 'block')
               runnable = [];
           elseif ~isempty(find(runnablesHdls == parent, 1))
               runnable = parent;
           else
               runnable = autosar.validation.ExportFcnValidator.findRunnableOwner(parent, runnablesHdls);
           end
       end

       function runnablesHdls = getRunnableHdls(ssBlkH, isMultiRunnable)
           
           runnablesHdls = [];
           if isMultiRunnable
               blks = autosar.validation.ExportFcnValidator.findSystemThruVirtualSubsystems(...
                   ssBlkH, 'block');
               for idx = 1:length(blks)
                   if strcmpi(get_param(blks(idx), 'BlockType'), 'SubSystem') && ...
                           strcmp(RTW.AutosarInterface.p_checkTrigSSTypeUsingTrigPortBlock(blks(idx)),'function-call')
                       runnablesHdls(end+1) = blks(idx); %#ok<AGROW>
                   end
               end
           end
           
       end
        
       function isMultiRunnable = isMultiRunnable(blkH)
       
           switch get_param(blkH, 'Type')
               case 'block_diagram'
                   isMultiRunnable = true;
               otherwise
                    isMultiRunnable = ~strcmp(RTW.AutosarInterface.p_checkTrigSSTypeUsingTrigPortBlock(blkH),'function-call');
           end
       end
       
    end
    
end

% LocalWords:  sobj Mdlref Goto's multirunnable froms SSRTW Unconfigured FCSS
% LocalWords:  Interrunnable noncomplex Irvs metamodel arplatform AUTOSARCS
