classdef SimulinkMappingValidator < autosar.validation.PhasedValidator
% SimulinkMappingValidator
    
%   Copyright 2013-2016 The MathWorks, Inc.

    methods (Access=protected)
        
        function verifyInitial(this, hModel)
            this.verifyMapping(hModel);
            this.verifyDuplicateCallerBlocks(hModel);
            this.verifySLPortRTEAPIs(hModel);
            this.verifySlFcnAR3xSupport(hModel)
            this.verifyNvPortDataAccessModes(hModel);
        end

        function verifyPostProp(this, hModel)
            this.verifySLPortStatusConnections(hModel);
            this.verifySampleTimesStepFunctions(hModel);
            this.verifyRateTransitions(hModel);
            this.verifySignalInvalidationBlocks(hModel);
            this.verifyRateTransitionDataTypes(hModel);
            this.verifySLPortElementDataTypes(hModel);
            this.verifyReferencedDataTypeMapping(hModel);
            this.verifyInterfaceQueuedConsistency(hModel)
            this.verifyLookupTables(hModel);
        end
        
    end

    methods (Static, Access=private)
        function verifyMapping(hModel)
            mapping = autosar.api.Utils.modelMapping(hModel);
            mapping.sync();
            mapping.validate();
        end
    
        function verifyReferencedDataTypeMapping(hModel)
            import Simulink.metamodel.types.CompuMethodCategory;
            import autosar.mm.util.ExternalToolInfoAdapter;
            
            mapping = autosar.api.Utils.modelMapping(hModel);
            m3iModel = autosar.api.Utils.m3iModel(hModel);
            arRoot = m3iModel.RootPackage.front();
            m3iSeq = Simulink.metamodel.arplatform.ModelFinder.findObjectByMetaClass(arRoot,...
                Simulink.metamodel.types.CompuMethod.MetaClass, true);
            for ii = 1 : m3iSeq.size()              
                compuMethod = m3iSeq.at(ii);
                slDataTypes = ExternalToolInfoAdapter.get(compuMethod,...
                    autosar.ui.metamodel.PackageString.SlDataTypes);
                for jj = 1 : numel(slDataTypes)
                    if ~isempty(compuMethod) && compuMethod.isvalid()
                        errorCodes = autosar.mm.util.checkDataTypeCompuMethodCompatibility(...
                            mapping.Name, slDataTypes{jj}, compuMethod, true);
                        if numel(errorCodes) > 0
                            DAStudio.error(errorCodes{1:numel(errorCodes)})
                        end
                    end
                end
            end
            if autosar.mm.arxml.Exporter.hasExternalReference(arRoot)
              m3iSeq = Simulink.metamodel.arplatform.ModelFinder.findObjectByParentMetaClass(arRoot,...
                Simulink.metamodel.foundation.ValueType.MetaClass, true);
              for ii = 1 : m3iSeq.size()
                implDataType = m3iSeq.at(ii);
                if ~implDataType.IsApplication
                  slDataTypes = ExternalToolInfoAdapter.get(implDataType,...
                    autosar.ui.metamodel.PackageString.SlDataTypes);
                  for jj = 1 : numel(slDataTypes)
                    if ~isempty(implDataType) && implDataType.isvalid()
                      errorCodes = autosar.mm.util.checkImplementationDataTypeCompatibility(...
                        mapping.Name, slDataTypes{jj}, implDataType);
                      if numel(errorCodes) > 0
                        DAStudio.error(errorCodes{1:numel(errorCodes)})
                      end
                    end
                  end
                end
              end
            end
        end
        
        function verifySLPortElementDataTypes(hModel)
           % Verify consistent data types for inports/outports
            
           mapping = autosar.api.Utils.modelMapping(hModel);
           dataObj = autosar.api.getAUTOSARProperties(hModel, true);
           componentQName = dataObj.get('XmlOptions', 'ComponentQualifiedName');

           % keep a map of arPort/arElement -> model's InPort Idx
           % This is the reverse of what is stored in mapping.Inports
           arPortToSlPortNumMap = containers.Map();
           for dataIdx = 1:length(mapping.Inports)
              inport = mapping.Inports(dataIdx); 
              arPortName = inport.MappedTo.Port;
              arElementName = inport.MappedTo.Element;
              switch inport.MappedTo.DataAccessMode
                case {'ImplicitReceive', ...
                      'ExplicitReceive', ...
                      'QueuedExplicitReceive', ...
                      'EndToEndRead'}
                  mapKey = [arPortName, '/', arElementName];
                  arPortToSlPortNumMap(mapKey) = dataIdx;
                otherwise
                  %do nothing
              end
           end
           
           % Verify data elements have consistent data types
           elementInfoMap = containers.Map();
           for inportIdx = 1:length(mapping.Inports)
               inport = mapping.Inports(inportIdx);
               elementInfoMap = autosar.validation.SimulinkMappingValidator.i_checkElement( dataObj, componentQName, ...
                                                                 elementInfoMap, inport, true);

               switch inport.MappedTo.DataAccessMode
                   case 'ExplicitReceiveByVal'
                       autosar.validation.SimulinkMappingValidator.i_checkExplicitReceiveByValDataType(hModel, inport);
                   case 'ErrorStatus'
                       arPortName = inport.MappedTo.Port;
                       arElementName = inport.MappedTo.Element;
                       mapKey = [arPortName, '/', arElementName];
                       receiverInPort = mapping.Inports(arPortToSlPortNumMap(mapKey));
                       isE2E = strcmp(receiverInPort.MappedTo.DataAccessMode, 'EndToEndRead');
                       autosar.validation.SimulinkMappingValidator.i_checkErrorStatusDataType(inport, isE2E);
                   case 'IsUpdated'
                       autosar.validation.SimulinkMappingValidator.i_checkIsUpdatedDataType(inport);
                   case 'ModeReceive'
                       autosar.validation.SimulinkMappingValidator.i_checkModeDataType(inport, hModel);
                   otherwise
                       % Do nothing.
               end
           end
           
           for outportIdx = 1:length(mapping.Outports)
               outport = mapping.Outports(outportIdx);
               elementInfoMap = autosar.validation.SimulinkMappingValidator.i_checkElement( dataObj, componentQName, ...
                                                                 elementInfoMap, outport, false);
               switch outport.MappedTo.DataAccessMode
                   case 'ModeSend'
                       autosar.validation.SimulinkMappingValidator.i_checkModeDataType(outport, hModel);
                   otherwise
                       % Do nothing.
               end                                                               
           end

        end
        
        function statusStr = convertBoolToEnabledStr(booleanValue)
            if booleanValue
                statusStr = 'enabled';
            else
                statusStr = 'disabled';
            end
        end
        
        function verifySLPortStatusConnections(hModel)
           % Verify error status and is updated connections
            
           mapping = autosar.api.Utils.modelMapping(hModel);

           % Verify only one error status or is updated per data port
           referredToByErrorStatus = {};
           referringToErrorStatus = {};
           referredToByIsUpdated = {};
           referringToIsUpdated = {};
           for statusInportIdx = 1:length(mapping.Inports)
               statusInport = mapping.Inports(statusInportIdx);
               statusDataAccessMode = statusInport.MappedTo.DataAccessMode;
               if ~ismember(statusDataAccessMode, {'ErrorStatus', 'IsUpdated'})
                   continue
               end
               statusInportName = get_param(statusInport.Block, 'Name');
               
               arPortName = statusInport.MappedTo.Port;
               arElementName = statusInport.MappedTo.Element;
               statusIsEnabled = strcmp(get_param(statusInport.Block, 'CompiledIsActive'), 'on');
               
               for dataIdx = 1:length(mapping.Inports)
                   dataInport = mapping.Inports(dataIdx);
                   dataInportName = get_param(dataInport.Block, 'Name');
                   dataAccessMode = dataInport.MappedTo.DataAccessMode;
                   
                   if ~ismember(dataAccessMode, {'ImplicitReceive', ...
                           'ExplicitReceive', 'ExplicitReceiveByVal', ...
                           'EndToEndRead', 'QueuedExplicitReceive', ...
                           })
                       continue
                   end
                               
                   dataInportIsEnabled = strcmp(...
                       get_param(dataInport.Block, 'CompiledIsActive'), ...
                       'on');
                   if strcmp(dataInport.MappedTo.Port, arPortName) && ...
                      strcmp(dataInport.MappedTo.Element, arElementName)
                       switch statusDataAccessMode
                           case {'ErrorStatus'}
                               if strcmp(dataAccessMode, 'ExplicitReceiveByVal')
                                   DAStudio.error('autosarstandard:validation:explicitReceiveByValErrorStatus', ...
                                                  statusInport.Block, ...
                                                  dataInport.Block);
                               end
                               existingRef = find(ismember(referredToByErrorStatus, ...
                                                           dataInportName));
                               if ~isempty(existingRef)
                                   msg = DAStudio.message('RTW:autosar:errorStatusPortNumClash', ...
                                                          statusInportName, ...
                                                          dataInportName, ...
                                                          referringToErrorStatus{existingRef}, ...
                                                          dataInportName);
                                   DAStudio.error('RTW:fcnClass:finish', msg);
                               end

                               if (statusIsEnabled ~= dataInportIsEnabled)
                                   statusEnabledStr = autosar.validation.SimulinkMappingValidator.convertBoolToEnabledStr(statusIsEnabled);
                                   dataEnabledStr = autosar.validation.SimulinkMappingValidator.convertBoolToEnabledStr(dataInportIsEnabled);
                                   DAStudio.error('autosarstandard:validation:errorStatusVariantMismatch', ...
                                       statusInport.Block, statusEnabledStr, ...
                                       dataInport.Block, dataEnabledStr);
                               end
                               
                               referredToByErrorStatus{end+1} = dataInportName; %#ok<AGROW>
                               referringToErrorStatus{end+1} = statusInportName; %#ok<AGROW>
                           case {'IsUpdated'}
                               existingRef = find(ismember(referredToByIsUpdated, ...
                                                           dataInportName));
                               if ~isempty(existingRef)
                                   DAStudio.error('autosarstandard:validation:isUpdatedPortNumClash', ...
                                                  statusInport.Block, ...
                                                  dataInport.Block, ...
                                                  referringToIsUpdated{existingRef});
                               end
                               referredToByIsUpdated{end+1} = dataInportName; %#ok<AGROW>
                               referringToIsUpdated{end+1} = statusInport.Block; %#ok<AGROW>
                               
                               if ~ismember(dataAccessMode, {...
                                       'ExplicitReceive', ...
                                       'ExplicitReceiveByVal', ...
                                       'EndToEndRead'})
                                   DAStudio.error('autosarstandard:validation:isUpdatedPortWithoutExplicitReceive', ...
                                       dataInport.Block, ...
                                       statusInport.Block);
                               end
                               
                               if (statusIsEnabled ~= dataInportIsEnabled)
                                   statusEnabledStr = autosar.validation.SimulinkMappingValidator.convertBoolToEnabledStr(statusIsEnabled);
                                   dataEnabledStr = autosar.validation.SimulinkMappingValidator.convertBoolToEnabledStr(dataInportIsEnabled);
                                   DAStudio.error('autosarstandard:validation:isUpdatedVariantMismatch', ...
                                       statusInport.Block, statusEnabledStr, ...
                                       dataInport.Block, dataEnabledStr);
                               end
                           otherwise
                              assert(false, 'Did not recognize DataAccessMode %s', statusDataAccessMode);
                       end
                       break
                   end
               end
               
               % check that error status port points to the same runnables as the corresponding data port
               if autosar.validation.ExportFcnValidator.isExportFcn(hModel)
                   dataPortH = get_param(dataInport.Block, 'Handle');
                   dataPortLineH = get_param(dataPortH, 'LineHandles');
                   dataPortLineObj = get_param(dataPortLineH.Outport, 'Object');
                   
                   statusPortH = get_param(statusInport.Block, 'Handle');
                   statusPortLineH = get_param(statusPortH, 'LineHandles');
                   statusPortLineObj = get_param(statusPortLineH.Outport, 'Object');
                   
                   dataPortDstBlocks = autosar.validation.walkThruVirtualBlockToDsts(dataPortLineObj, hModel, true);
                   statusPortDstBlocks = autosar.validation.walkThruVirtualBlockToDsts(statusPortLineObj, hModel, true);
                   
                   % Work through error port destinations blocks to see which are runnables
                   for dstIdx=1:length(statusPortDstBlocks)
                       if strcmp(RTW.AutosarInterface.p_checkTrigSSTypeUsingTrigPortBlock(statusPortDstBlocks(dstIdx)),'function-call')
                           
                           % check if runnable is also a destination of the data port
                           if ~ismember(statusPortDstBlocks(dstIdx), dataPortDstBlocks)
                               switch statusDataAccessMode
                                   case {'ErrorStatus'}
                                       invalidRunnableErrorId = 'autosarstandard:validation:errorStatusPortInvalidRunnable';
                                   case {'IsUpdated'}
                                       invalidRunnableErrorId = 'autosarstandard:validation:isUpdatedPortInvalidRunnable';
                                   otherwise
                                       assert(false, 'Did not recognize DataAccessMode %s', statusDataAccessMode);
                               end
                               DAStudio.error(invalidRunnableErrorId, ...
                                              dataInport.Block, ...
                                              statusInport.Block);
                           end
                       end
                   end
               end
               
           end

       end
       
       function verifySampleTimesStepFunctions(hModel)
           if autosar.validation.ExportFcnValidator.isExportFcn(hModel)
               return
           end
           
           mapping = autosar.api.Utils.modelMapping(hModel);
           mapping.validateStepFunctions()
       end
       
       function verifyRateTransitions(hModel)
           mapping = autosar.api.Utils.modelMapping(hModel);
           if autosar.validation.ExportFcnValidator.isExportFcn(hModel)
               % Skip validation for export function models.               return
           end
                      
           if ~isempty(mapping.RateTransition)
               dataObj = autosar.api.getAUTOSARProperties(hModel, true);
               componentQualifiedName = dataObj.get('XmlOptions', ...
                   'ComponentQualifiedName');
               
               irvNames = arrayfun(@(x) x.MappedTo.IrvName, ...
                   mapping.RateTransition, 'UniformOutput', false);
               for i = 1: length(mapping.RateTransition)
                   if isempty(mapping.RateTransition(i).MappedTo.IrvName)
                       blkPath = mapping.RateTransition(i).Block;
                       DAStudio.error(...
                           'autosarstandard:validation:unmappedRateTransition',...
                           blkPath);
                   else
                       if any(strcmp(irvNames(i),irvNames(i+1:end)))
                           blkPath = mapping.RateTransition(i).Block;
                           DAStudio.error(...
                               'autosarstandard:validation:duplicateIRVMappedToRTB',...
                               blkPath, irvNames{i});
                       else
                           ARIRVPath = dataObj.find(componentQualifiedName, ...
                               'IrvData', 'Name', irvNames{i});
                           if isempty(ARIRVPath)
                               % Possible only if IRV mapping was created
                               % via command-line
                               DAStudio.error(...
                                   'autosarstandard:validation:irvDoesNotExist',...
                                   mapping.RateTransition(i).Block,...
                                   irvNames{i});
                           end
                       end
                   end
               end            
           end
           
           if length(mapping.StepFunctions) <= 1
               % Skip validation for single-tasking models.
               return;
           end
           
           rtBlks = find_system(get_param(hModel, 'Name'), 'FollowLinks', 'on',...
               'LookUnderMasks', 'all', 'BlockType', 'RateTransition');
           for i = 1:length(rtBlks)
               if ~(strcmp(get_param(rtBlks{i},'Integrity'), 'on') && ...
                       strcmp(get_param(rtBlks{i},'Deterministic'), 'off'))
                   DAStudio.error(...
                       'autosarstandard:validation:RateTranBlkForIRV',...
                       rtBlks{i});
               end             
           end
       end
       
       function verifyRateTransitionDataTypes(hModel)
           mapping = autosar.api.Utils.modelMapping(hModel);
           if autosar.validation.ExportFcnValidator.isExportFcn(hModel) || ...
                   (length(mapping.StepFunctions) <= 1)
               % Skip validation for export function models and
               % single-tasking models
               return
           end
           
           if ~isempty(mapping.RateTransition)
               cs = autosar.validation.ConfigSetValidator.getActiveConfigSet(hModel);
               maxShortNameLength = get_param(cs, 'AutosarMaxShortNameLength');
               
               isSupportMatrixIOAsArray = strcmpi(...
                   get_param(cs,'AutosarMatrixIOAsArray'), 'on');               
               
               for i = 1: length(mapping.RateTransition)
                   blk = mapping.RateTransition(i).Block;
                   % check that it is a scalar noncomplex
                   complexSigs = get_param(blk,'CompiledPortComplexSignals');
                   if complexSigs.Inport
                       DAStudio.error('autosarstandard:validation:RTBComplex', ...
                                      blk);
                   end
                   
                   % Only checks if IRV is array or struct if it is pre
                   % AUTOSAR 4.0
                   if ~AR4.isCompatibleSchemaVer( ...
                           autosar.SchemaVersionHandler.instance().getSchemaVersion() )
                       portWidths = get_param(blk,'CompiledPortWidths');
                       if portWidths.Inport > 1
                           DAStudio.error('autosarstandard:validation:RTBArray', ...
                                          blk);        
                       end
                       
                       dataTypes = get_param(blk,'CompiledPortDataTypes');                       
                       if existsInGlobalScope(hModel, dataTypes.Inport{1}) ...
                           && evalinGlobalScope(hModel, ...
                           ['isa(' dataTypes.Inport{1} ',''Simulink.Bus'')'])
                           DAStudio.error('autosarstandard:validation:RTBBus', ...
                                          blk);
                       else
                           % For AUTOSAR 4.0 check if the datatype is valid.
                           autosar.validation.AutosarUtils.checkDataType(...
                               hModel, blk, dataTypes.Inport{1}, ...
                               maxShortNameLength,isSupportMatrixIOAsArray)
                       end                   
                   end
               end               
           end
       end       
       
       function verifySLPortRTEAPIs(hModel)
           
           mapping = autosar.api.Utils.modelMapping(hModel);
           
           % Verify APIs don't clash
           apiMap = containers.Map();
           for inportIdx = 1:length(mapping.Inports)
               inport = mapping.Inports(inportIdx);
               apiMap = autosar.validation.SimulinkMappingValidator.i_checkAPI( hModel, apiMap, inport);
           end
           
           for outportIdx = 1:length(mapping.Outports)
               outport = mapping.Outports(outportIdx);
               apiMap = autosar.validation.SimulinkMappingValidator.i_checkAPI( hModel, apiMap, outport);
           end
           
       end
       
       function verifyDuplicateCallerBlocks(hModel)
           mapping = autosar.api.Utils.modelMapping(hModel);
           dups = containers.Map;
           for index=1:length(mapping.FunctionCallers)
               if ~isempty(mapping.FunctionCallers(index).MappedTo)
                   key = [mapping.FunctionCallers(index).MappedTo.ClientPort '/' ...
                       mapping.FunctionCallers(index).MappedTo.Operation];
                   if ~isKey(dups, key)
                       dups(key) = mapping.FunctionCallers(index).Block;
                   else
                       if ~strcmp(get_param(mapping.FunctionCallers(index).Block, 'FunctionPrototype'),...
                               get_param(dups(key), 'FunctionPrototype'))
                           msg = DAStudio.message('RTW:autosar:duplicateMappedClientBlock', ...
                               mapping.FunctionCallers(index).Block, ...
                               dups(key));
                           DAStudio.error('RTW:fcnClass:finish', msg);
                       end
                   end
               end
           end
       end
       

       function verifySlFcnAR3xSupport(hModel)
       % Check Simulink Function AR3x support. NvM Service feature is only
       % valid for 4.x

           doCheckForNvm = slfeature('AUTOSARNvmRwCg')>0;  % NvM is never supported so throw error if enabled
           doCheckForArrSizeOne = slfeature('AUTOSARArrSizeOneForAR3x')<=0;
           if ~(doCheckForNvm || doCheckForArrSizeOne)
               return;
           end
       
           if ~AR4.isCompatibleSchemaVer(autosar.SchemaVersionHandler.instance().getSchemaVersion)
               modelMapping = autosar.api.Utils.modelMapping(hModel);
               for ii=1:length(modelMapping.FunctionCallers)
                   fcnCaller = modelMapping.FunctionCallers(ii);
                   if ~isempty(fcnCaller)
                       m3iOperation = ...
                           autosar.validation.ClientServerValidator.findM3iOpFromPortOpName(hModel, ...
                           fcnCaller.MappedTo.ClientPort, ...
                           fcnCaller.MappedTo.Operation);
                       
                       if(doCheckForNvm)
                           if autosar.validation.ClientServerValidator.isNvMService(m3iOperation)
                               DAStudio.error('autosarstandard:validation:invalidNvmServiceFcnCaller', ...
                                   fcnCaller.Block);
                           end
                       end
                       if(doCheckForArrSizeOne)
                           for j=1:m3iOperation.Arguments.size
                               if(autosar.api.Utils.isArraySizeOne(m3iOperation.Arguments.at(j).Type))
                                   DAStudio.error('autosarstandard:validation:arraySizeOneUnsupportedFcnCaller', ...
                                       fcnCaller.Block,j);
                               end
                           end
                       end
                   end
               end
               
               for ii=1:length(modelMapping.ServerFunctions)
                   slFunction = modelMapping.ServerFunctions(ii);
                   if ~isempty(slFunction)
                       m3iRunnable = autosar.validation.ClientServerValidator.findM3iRunnableFromName( ...
                           hModel, slFunction.MappedTo.Runnable);
                       allM3IOp = autosar.validation.ClientServerValidator.getAllRunnableOperations(m3iRunnable);
                       assert(length(allM3IOp) >= 1, 'Expected one or more operations for runnable %s', m3iRunnable.Name);
                       m3iOperation = allM3IOp{1};
                       
                       if(doCheckForNvm)
                           if autosar.validation.ClientServerValidator.isNvMService(m3iOperation)
                               DAStudio.error('autosarstandard:validation:invalidNvmServiceSLFcn', ...
                                   slFunction.Block);
                           end
                       end
                       if(doCheckForArrSizeOne)
                           for j=1:m3iOperation.Arguments.size
                               if(autosar.api.Utils.isArraySizeOne(m3iOperation.Arguments.at(j).Type))
                                   DAStudio.error('autosarstandard:validation:arraySizeOneUnsupportedSLFcn', ...
                                       slFunction.Block,j);
                               end
                           end
                       end

                   end
               end
               
           end
           
       end

       function verifyNvPortDataAccessModes(hModel)
       % Verify data access modes for nonvolatile ports
           
           isAR4 = AR4.isCompatibleSchemaVer( ...
               autosar.SchemaVersionHandler.instance().getSchemaVersion() );

           if ~isAR4
               return
           end
           
           dataObj = autosar.api.getAUTOSARProperties(hModel, true);
           compQName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
           nvReceiverPorts = dataObj.get(compQName, 'NvReceiverPorts', ...
                                          'PathType', 'FullyQualified');
           nvSenderPorts = dataObj.get(compQName, 'NvSenderPorts', ...
                                          'PathType', 'FullyQualified');
           nvSenderReceiverPorts = dataObj.get(compQName, 'NvSenderReceiverPorts', ...
                                                'PathType', 'FullyQualified');

           arPorts = [nvReceiverPorts, nvSenderPorts, nvSenderReceiverPorts];
           if isempty(arPorts)
               return
           end

           % Just need port names
           for ii=1:length(arPorts)
              arPorts{ii} = dataObj.get(arPorts{ii}, 'Name'); 
           end
           
           % Check data access modes via mapping
           mapping = autosar.api.Utils.modelMapping(hModel);
           for dataIdx = 1:length(mapping.Inports)
               blkMap = mapping.Inports(dataIdx); 
               dataAccessMode = blkMap.MappedTo.DataAccessMode;
               switch dataAccessMode
                 case autosar.api.getSimulinkMapping.getValidNvReceiverDAMs()
                   % Skip if valid
                 otherwise
                   arPortName = blkMap.MappedTo.Port;
                   arElementName = blkMap.MappedTo.Element;
                   if any(strcmp(arPortName, arPorts))
                       DAStudio.error('autosarstandard:validation:invalidNvPortDAM', ...
                                      blkMap.Block, arPortName, arElementName, dataAccessMode);
                   end
               end
           end
           for dataIdx = 1:length(mapping.Outports)
               blkMap = mapping.Outports(dataIdx); 
               dataAccessMode = blkMap.MappedTo.DataAccessMode;
               switch dataAccessMode
                 case autosar.api.getSimulinkMapping.getValidNvSenderDAMs()
                   % Skip if valid
                 otherwise
                   arPortName = blkMap.MappedTo.Port;
                   arElementName = blkMap.MappedTo.Element;
                   if any(strcmp(arPortName, arPorts))
                       DAStudio.error('autosarstandard:validation:invalidNvPortDAM', ...
                                      blkMap.Block, arPortName, arElementName, dataAccessMode);
                   end
               end
           end

       end
       
       
       function verifySignalInvalidationBlocks(hModel)
           siBlks = find_system(get_param(hModel, 'Name'), ...
               'FollowLinks', 'on', 'LookUnderMasks', 'all', ...
               'BlockType', 'SignalInvalidation');
           autosarVer = autosar.SchemaVersionHandler.instance().getSchemaVersion();
           for i = 1:length(siBlks)
               currBlk = siBlks{i};
               
               cLine = get_param(currBlk, 'LineHandles');
               currOutport = cLine.Outport;
               
               if currOutport < 0
                   DAStudio.error('autosarstandard:validation:signalInvalidationBlockActualDestination', ...
                       currBlk);
               end
               
               dataPortLineObj = get_param(currOutport, 'Object');
               assert(dataPortLineObj.SrcPortHandle >= 0);
               srcPortObj = get_param(dataPortLineObj.SrcPortHandle,'Object');
               
               % g1262543: Eliminate the slfeature commands once the 
               % EngineInterface is enabled by default.
               eiPval = slfeature('EngineInterface', 1); 
               actDstInfo = srcPortObj.getActualDst(); 
               slfeature('EngineInterface', eiPval); 
                
               if (size(actDstInfo,1) < 1) 
                   DAStudio.error('autosarstandard:validation:signalInvalidationBlockActualDestination', ...
                       currBlk);
               elseif (size(actDstInfo,1) > 1) 
                   DAStudio.error('autosarstandard:validation:signalInvalidationBlockMultipleDestinations', ...
                       currBlk);
               end 
               dstPortHdl = actDstInfo(1); 
               outBlk = get_param(dstPortHdl, 'Parent'); 

               if ~AR4.isCompatibleSchemaVer(autosarVer) && ...
                  strcmp(get_param(outBlk, 'BlockType'), 'SignalConversion')
                   DAStudio.error('autosarstandard:validation:signalInvalidationBlockActualDestinationHiddenBuffer', ...
                       currBlk);
               end
               
               if ~strcmp(get_param(outBlk, 'BlockType'), 'Outport') || ... 
                  (hModel ~= get_param(get_param(outBlk, 'Parent'), 'Handle')) 
                   DAStudio.error('autosarstandard:validation:signalInvalidationBlockActualDestination', ...
                       currBlk);
               end 
               
               outBlkName = get_param(outBlk, 'Name');
               outBlkFullName = getfullname(outBlk);
               
               mapping = autosar.api.getSimulinkMapping(hModel);
               [outBlkPort, outBlkElem, outBlkDataAccess] = mapping.getOutport(outBlkName);
               if ~ismember(outBlkDataAccess, ...
                       {'ExplicitSend', 'EndToEndWrite'})
                   DAStudio.error('autosarstandard:validation:signalInvalidationBlockAccessMode', ...
                       outBlkFullName, currBlk);
               end
               [invPolicy, sigInvInterface] = autosar.validation.AutosarUtils.getInterfaceElementProperty(...
                   hModel, outBlkPort, outBlkElem, 'InvalidationPolicy');
               if ismember(invPolicy, {'None', 'DontInvalidate'})
                   DAStudio.error('autosarstandard:validation:signalInvalidationBlockPolicy', ...
                       outBlkFullName, currBlk, outBlkElem, sigInvInterface, invPolicy);
               end
           end
       end
       
       function verifyInterfaceQueuedConsistency(hModel)
           % AR 3.X/2.X check only: Verify that the queued attribute for
           % two inports are the same when they are mapped to the same
           % interface data element.
           
           autosarVer = autosar.SchemaVersionHandler.instance().getSchemaVersion();
           if ~AR4.isCompatibleSchemaVer(autosarVer)
               mapping = autosar.api.Utils.modelMapping(hModel);
               dataObj = autosar.api.getAUTOSARProperties(hModel, true);
               componentQName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
               
               % Keep a map of arInterface/arElement -> Inport.
               arIfDeToInportMap = containers.Map();
               for dataIdx = 1:length(mapping.Inports)
                   inport = mapping.Inports(dataIdx);
                   arPortName = inport.MappedTo.Port;
                   arElementName = inport.MappedTo.Element;
                   dataAccessMode = inport.MappedTo.DataAccessMode;
                   switch dataAccessMode
                       case {'ExplicitReceive', 'ExplicitReceiveByVal', ...
                               'EndToEndRead'}
                           isQueued = false;
                       case {'QueuedExplicitReceive'}
                           isQueued = true;
                       otherwise
                           continue;
                   end
                   interfaceQName = dataObj.get(...
                       [componentQName, '/', arPortName], ...
                       'Interface', 'PathType', 'FullyQualified');
                   elementQName = [interfaceQName, '/' arElementName];
                   
                   if ~arIfDeToInportMap.isKey(elementQName)
                       arIfDeToInportMap(elementQName) = inport;
                   else
                       otherInport = arIfDeToInportMap(elementQName);
                       otherDataAccessMode = otherInport.MappedTo.DataAccessMode;
                       % We know that the other inport is one of the
                       % expected access mode. Hence, we only compare it
                       % against QueuedExplicitReceive.
                       otherIsQueued = strcmp(otherDataAccessMode, ...
                           'QueuedExplicitReceive');
                       if isQueued ~= otherIsQueued
                           autosar.validation.SimulinkMappingValidator.i_queuedAttributeThroughout(...
                               dataObj, componentQName, interfaceQName, ...
                               arElementName, arPortName, isQueued, ...
                               otherInport);
                       end
                   end
               end
               if isempty(arIfDeToInportMap)
                   % We do not add outports to the map because an
                   % outport will never have a queued configuration
                   % conflict with another. Hence, if the map is empty
                   % there is nothing else to be checked.
                   return;
               end
               for dataIdx = 1:length(mapping.Outports)
                   outport = mapping.Outports(dataIdx);
                   arPortName = outport.MappedTo.Port;
                   arElementName = outport.MappedTo.Element;
                   % A root outport can never be queued.
                   isQueued = false;
                   interfaceQName = dataObj.get(...
                       [componentQName, '/', arPortName], ...
                       'Interface', 'PathType', 'FullyQualified');
                   elementQName = [interfaceQName, '/' arElementName];
                   if arIfDeToInportMap.isKey(elementQName)
                       otherInport = arIfDeToInportMap(elementQName);
                       otherDataAccessMode = otherInport.MappedTo.DataAccessMode;
                       % We know that the other inport is one of the
                       % expected access mode. Hence, we only compare it
                       % against QueuedExplicitReceive.
                       otherIsQueued = strcmp(otherDataAccessMode, ...
                           'QueuedExplicitReceive');
                       if isQueued ~= otherIsQueued
                           autosar.validation.SimulinkMappingValidator.i_queuedAttributeThroughout(...
                               dataObj, componentQName, interfaceQName, ...
                               arElementName, arPortName, isQueued, ...
                               otherInport);
                       end
                   end
               end
           end
       end
       
       function verifyLookupTables(hModel)
           % Verify that Simulink lookup tables are not mapped to same
           % AUTOSAR Parameter. For example Simulink.LookupTable with names
           % lut1 and lut2 cannot be mapped to same AUTOSAR parameter P1.
           modelMapping = autosar.api.Utils.modelMapping(hModel);
           dataObj = autosar.api.getAUTOSARProperties(hModel, true);
           componentQualifiedName = dataObj.get('XmlOptions', ...
               'ComponentQualifiedName');
           m3iModel = autosar.api.Utils.m3iModel(hModel);
           compObjSeq = autosar.mm.Model.findObjectByName(m3iModel, componentQualifiedName);
           assert(compObjSeq.size() > 0);
           compObj = compObjSeq.at(1);
           paramsMap = containers.Map;
           for ii = 1 : numel(modelMapping.LookupTables)
               mapping = modelMapping.LookupTables(ii);
               if strcmp(mapping.MappedTo.ParameterAccessMode, 'PortParameter')
                   if paramsMap.isKey([mapping.MappedTo.Port '.' mapping.MappedTo.Parameter])
                       ndx = paramsMap([mapping.MappedTo.Port '.' mapping.MappedTo.Parameter]);
                       param1 = modelMapping.LookupTables(ndx).LookupTableName;
                       param2 = mapping.LookupTableName;
                       if existsInGlobalScope(hModel, param1) &&...
                               existsInGlobalScope(hModel, param2)
                           obj1 = evalinGlobalScope(hModel, param1);
                           obj2 = evalinGlobalScope(hModel, param2);
                           DAStudio.error('autosarstandard:validation:duplicateMappedParameters',...
                               class(obj1), param1, class(obj2), param2);
                       end
                   else
                       paramsMap([mapping.MappedTo.Port '.' mapping.MappedTo.Parameter]) = ii;
                   end
                   m3iPort = Simulink.metamodel.arplatform.ModelFinder.findNamedItemInSequence(...
                       compObj, compObj.ParameterReceiverPorts, mapping.MappedTo.Port,...
                       'Simulink.metamodel.arplatform.port.ParameterReceiverPort');
                   assert(m3iPort.isvalid(), 'Invalid Parameter Receiver Port.');
                   m3iData = Simulink.metamodel.arplatform.ModelFinder.findNamedItemInSequence(m3iPort.Interface,...
                       m3iPort.Interface.DataElements, mapping.MappedTo.Parameter,...
                       'Simulink.metamodel.arplatform.interface.ParameterData');
                   if ~m3iData.isvalid()
                        DAStudio.error(...
                           'autosarstandard:validation:ParamDoesNotExist',...
                           mapping.LookupTableName, mapping.MappedTo.Parameter,...
                           mapping.MappedTo.ParameterAccessMode);
                   end                   
               else
                   if paramsMap.isKey(mapping.MappedTo.Parameter)
                       ndx = paramsMap(mapping.MappedTo.Parameter);
                       param1 = modelMapping.LookupTables(ndx).LookupTableName;
                       param2 = mapping.LookupTableName;
                       if existsInGlobalScope(hModel, param1) &&...
                               existsInGlobalScope(hModel, param2)
                           obj1 = evalinGlobalScope(hModel, param1);
                           obj2 = evalinGlobalScope(hModel, param2);
                           DAStudio.error('autosarstandard:validation:duplicateMappedParametersInternal',...
                               class(obj1), param1, class(obj2), param2, mapping.MappedTo.Parameter);
                       end
                   else
                       paramsMap(mapping.MappedTo.Parameter) = ii;
                   end
                   assert(~isempty(mapping.MappedTo.Parameter), 'Lookup Tables must be mapped to AUTOSAR Parameters.');                
                   behaviorObj = compObj.Behavior;
                   calPrmName =  mapping.MappedTo.Parameter;
                   m3iData = Simulink.metamodel.arplatform.ModelFinder.findNamedItemInSequence(...
                                compObj, behaviorObj.Parameters, calPrmName,...
                                'Simulink.metamodel.arplatform.interface.ParameterData');                  
                   if ~m3iData.isvalid()
                        DAStudio.error(...
                           'autosarstandard:validation:ParamDoesNotExist',...
                           mapping.LookupTableName, mapping.MappedTo.Parameter,...
                           mapping.MappedTo.ParameterAccessMode);
                   end
               end
           end
       end
       
       function i_queuedAttributeThroughout(dataObj, componentQName, interfaceQName, ...
               arElementName, arPortName, isQueued, otherPort)
           % This function dedicates to fetch parameters that the 
           % "queuedAttributeThroughout" error message requires.
           
           interfaceCategory = dataObj.get(...
               [componentQName, '/', arPortName], ...
               'Category');
           switch interfaceCategory
               case {'DataReceiverPort', 'DataSenderPort'}
                   InterfaceType = 'Sender-Receiver';
               case {'ClientPort', 'ServerPort'}
                   InterfaceType = 'Client-Server';
               case {'NvDataReceiverPort', 'NvDataSenderPort'}
                   InterfaceType = 'NVM';
               otherwise
                   InterfaceType = '';
           end
           msgPortName = DAStudio.message('RTW:autosar:simulinkPort', ...
               arPortName);
           msgOtherPortName = DAStudio.message('RTW:autosar:simulinkPort', ...
               otherPort.MappedTo.Port);
           errMsgHeader = DAStudio.message('RTW:autosar:incompatibleDataElement',...
               arElementName, InterfaceType, ...
               msgPortName, msgOtherPortName, interfaceQName);
           if isQueued
               isQueuedStrValue = 'TRUE';
               currIsQueuedStrValue = 'FALSE';
           else
               isQueuedStrValue = 'FALSE';
               currIsQueuedStrValue = 'TRUE';
           end
           msg = DAStudio.message('RTW:autosar:queuedAttributeThroughout', ...
               errMsgHeader, ...
               isQueuedStrValue, currIsQueuedStrValue);
           DAStudio.error('RTW:fcnClass:finish', msg);
       end
       
       function elementInfoMap = i_checkElement( dataObj, componentQName, elementInfoMap, port, isInport)

           dataAccessMode = port.MappedTo.DataAccessMode;
           switch dataAccessMode
             case {'ImplicitReceive', ...
                   'ExplicitReceive', ...
                   'ExplicitReceiveByVal', ...
                   'QueuedExplicitReceive', ...
                   'ImplicitSend', ...
                   'ImplicitSendByRef', ...
                   'ExplicitSend',...
                   'ModeReceive', ...
                   'EndToEndRead', ...
                   'EndToEndWrite',...
                   'ModeSend'}

               arPortName    = port.MappedTo.Port;
               arElementName = port.MappedTo.Element;
               slPortName    = get_param(port.Block, 'Name');
               
               interfaceQName = dataObj.get([componentQName, '/', arPortName], 'Interface', 'PathType', 'FullyQualified');
               elementQName = [interfaceQName, '/' arElementName];
               
               autosarVer = autosar.SchemaVersionHandler.instance().getSchemaVersion();
               if AR4.isCompatibleSchemaVer(autosarVer) && ~ismember(dataAccessMode, {'ModeReceive', 'ModeSend'})
                   displayFormat = dataObj.get(elementQName, 'DisplayFormat');
                   [isValid, msg] = autosar.validation.AutosarUtils.checkDisplayFormat(displayFormat, elementQName);
                    if ~isValid
                        DAStudio.error('RTW:fcnClass:finish', msg);
                    end
               end

               elementDataTypesAliased = get_param(port.Block, 'CompiledPortDataTypes');
               elementWidths = get_param(port.Block, 'CompiledPortWidths');
               if isInport
                   elementDataType = elementDataTypesAliased.Outport{1};
                   elementWidth = elementWidths.Outport(1);
               else
                   elementDataType = elementDataTypesAliased.Inport{1};
                   elementWidth = elementWidths.Inport(1);
               end
               
               elementDataType = autosar.validation.SimulinkMappingValidator.convertToAutosarPlatformTypeName(elementDataType);
               
               if elementInfoMap.isKey(elementQName)
                   elementInfo = elementInfoMap(elementQName);

                   usrReadableRefStr = DAStudio.message('RTW:autosar:simulinkPort', slPortName);
                   if ismember(dataAccessMode, {'ModeReceive', 'ModeSend'})
                       interfaceName = 'Mode-Switch';
                   else
                       interfaceName = 'Sender-Receiver';
                   end
                   errMsgHeader = DAStudio.message('RTW:autosar:incompatibleDataElement',...
                                                   arElementName, interfaceName, usrReadableRefStr, elementInfo.UsrReadableRefStr, interfaceQName );
                   
                   if ~strcmp(elementDataType, elementInfo.DataType)
                       msg = DAStudio.message('RTW:autosar:dataTypeThroughout', errMsgHeader, elementDataType, elementInfo.DataType);
                       DAStudio.error('RTW:fcnClass:finish', msg);
                   end
                   
                   if elementWidth ~= elementInfo.Width
                       msg = DAStudio.message('RTW:autosar:widthThroughout', errMsgHeader, elementWidth, elementInfo.Width);
                       DAStudio.error('RTW:fcnClass:finish', msg);
                   end
                   
               else
                   % Store info for next time
                   elementInfo.DataType = elementDataType;
                   elementInfo.Width    = elementWidth;
                   elementInfo.UsrReadableRefStr = DAStudio.message('RTW:autosar:simulinkPort', slPortName);
                   elementInfoMap(elementQName) = elementInfo;
               end
               
             case {'ErrorStatus', 'IsUpdated'}
               % No element check needed for these categories.
             otherwise
               assert(false, 'Did not recognize DataAccessMode %s', dataAccessMode);
           end

       end

       function i_checkExplicitReceiveByValDataType( hModel, port )
       % Check explicit receive by value data type.
           
           dataTypes = get_param(port.Block, 'CompiledPortAliasedThruDataTypes');
           dataType = dataTypes.Outport{1};
           
           if existsInGlobalScope(hModel, dataType) ...
                   && evalinGlobalScope(hModel, ...
                   ['isa(' dataType ',''Simulink.Bus'')'])
               DAStudio.error('autosarstandard:validation:explicitReceiveByValInvalidType', port.Block);
           end
           
           widths = get_param(port.Block, 'CompiledPortWidths');
           width = widths.Outport(1);

           if width > 1
               DAStudio.error('autosarstandard:validation:explicitReceiveByValInvalidType', port.Block);
           end
           
           complexity = get_param(port.Block, 'SignalType');
           if strcmp(complexity, 'complex')
               DAStudio.error('autosarstandard:validation:explicitReceiveByValInvalidType', port.Block);
           end
       end
       
       function i_checkErrorStatusDataType( port, isE2E )
       % check error status data type

           slPortName = get_param(port.Block, 'Name');
           
           dataTypes = get_param(port.Block, 'CompiledPortAliasedThruDataTypes');
           dataType = dataTypes.Outport{1};
           
           widths = get_param(port.Block, 'CompiledPortWidths');
           width = widths.Outport(1);
           
           if isE2E
               if ~strcmp(dataType, 'uint32')
                   msg = DAStudio.message('RTW:autosar:errorStatusPortInvalidDataType', slPortName, 'uint32');
                   DAStudio.error('RTW:fcnClass:finish', msg);
               end
           else
               if ~strcmp(dataType, 'uint8')
                   msg = DAStudio.message('RTW:autosar:errorStatusPortInvalidDataType', slPortName, 'uint8');
                   DAStudio.error('RTW:fcnClass:finish', msg);
               end
           end

           if width > 1
               msg = DAStudio.message('RTW:autosar:errorStatusPortInvalidWidth', slPortName);
               DAStudio.error('RTW:fcnClass:finish', msg);
           end
       end

       function i_checkIsUpdatedDataType( port )
       % check error status data type

           % IsUpdated is supported for AUTOSAR 4.x and 3.2
           autosarVer = autosar.SchemaVersionHandler.instance().getSchemaVersion();
           if ~(str2double(autosarVer) >= 3.2)
               DAStudio.error('autosarstandard:validation:isUpdatedAutosarVersion');
           end
           
           dataTypes = get_param(port.Block, 'CompiledPortAliasedThruDataTypes');
           dataType = dataTypes.Outport{1};
           
           widths = get_param(port.Block, 'CompiledPortWidths');
           width = widths.Outport(1);
           
           if ~strcmp(dataType, 'boolean')
               DAStudio.error('autosarstandard:validation:isUpdatedPortInvalidDataType', port.Block);
           end

           if width > 1
               DAStudio.error('autosarstandard:validation:isUpdatedPortInvalidDataType', port.Block);
           end
       end
       
       function i_checkModeDataType( port, hModel )
         %common function to check datatype for Mode Send and Receive
         % check data type
         %inport.MappedTo.DataAccessMode
         slPortName = get_param(port.Block, 'Name');
         
         dataTypes = get_param(port.Block, 'CompiledPortAliasedThruDataTypes');
         widths = get_param(port.Block, 'CompiledPortWidths');
         
         dataAccessMode = port.MappedTo.DataAccessMode;
         switch dataAccessMode
           case {'ModeReceive'}
             dataType = dataTypes.Outport{1};
             width = widths.Outport(1);
             
           case {'ModeSend'}
             dataType = dataTypes.Inport{1};
             width = widths.Inport(1);
             
           otherwise
             assert(false, 'DataAccessMode %s not recognized in Mode Send/Receive port validation', dataAccessMode);
         end
         
         if Simulink.data.isSupportedEnumClass(dataType) == 0
           DAStudio.error('autosarstandard:validation:invalidDataTypeForMode', slPortName);
         else
           schemaVersion = autosar.SchemaVersionHandler.instance().getSchemaVersion();
           if slfeature('ControlStorageTypeForEnums') == 1
             storageType = Simulink.data.getEnumTypeInfo(dataType, 'StorageType');
             if ~strcmp(storageType, 'int')
               if AR4.isCompatibleSchemaVer(schemaVersion)
                 MdgSupportedStorageTypes = {'uint8', 'uint16', 'int8', 'int16', 'int32'};
                 MdgSupportedStorageTypesStr = {'uint8, uint16, int8, int16, int32'};
               else
                 MdgSupportedStorageTypes = {'uint8', 'uint16'};
                 MdgSupportedStorageTypesStr = {'uint8, uint16'};
               end
               if ~any(strcmp(storageType, MdgSupportedStorageTypes))
                 DAStudio.error('autosarstandard:validation:invalidStorageTypeForMode',...
                   storageType, slPortName, schemaVersion,...
                   MdgSupportedStorageTypesStr{1});
               end
             end

             %Mode value should be non-negative for AR4, otherwise schema check fails
             if AR4.isCompatibleSchemaVer(schemaVersion)
                 [~, modeValues, ~, storageType, ~, ~] =...
                     autosar.mm.sl2mm.ModelBuilder.getMdgDataFromEnum(...
                         hModel, dataType, AR4.isCompatibleSchemaVer(schemaVersion));
                 uint8_min = intmin('uint8');
                 uint_max = intmax(storageType);
                 [minV, minIndex] = min(modeValues);
                 if minV < uint8_min
                     DAStudio.error('autosarstandard:validation:invalidModeValueForEnumeration',...
                                    dataType, modeValues(minIndex), slPortName,...
                                    schemaVersion, uint8_min, uint_max);
                 end
             end
             
           else
             [~, modeValues, ~, storageType, ~, ~] =...
               autosar.mm.sl2mm.ModelBuilder.getMdgDataFromEnum(...
                   hModel, dataType, AR4.isCompatibleSchemaVer(schemaVersion));
             if ~strcmp(storageType, 'int32')
               DAStudio.error( 'autosarstandard:validation:invalidEnumerationTypeForMode', ...
                               dataType, slPortName);
             end
             if ~AR4.isCompatibleSchemaVer(schemaVersion)
               uint16_min = intmin('uint16');
               uint16_max = intmax('uint16');
               [minV, minIndex] = min(modeValues);
               [maxV, maxIndex] = max(modeValues);
               if minV < uint16_min
                 DAStudio.error('autosarstandard:validation:invalidModeValueForEnumeration',...
                   dataType, modeValues(minIndex), slPortName,...
                   schemaVersion, uint16_min, uint16_max);
               end
               if maxV > uint16_max
                 DAStudio.error('autosarstandard:validation:invalidModeValueForEnumeration',...
                   dataType, modeValues(maxIndex), slPortName,...
                   schemaVersion, uint16_min, uint16_max);
               end
             end
           end
         end
         if width > 1
           DAStudio.error('autosarstandard:validation:modePortInvalidWidth', ...
                          slPortName);
         end
       end
       
       
       function apiMap = i_checkAPI( hModel, apiMap, port)
       % Check for API clashes
           
           arPortName    = port.MappedTo.Port;
           arElementName = port.MappedTo.Element;
           arDataAccessMode = port.MappedTo.DataAccessMode;
           autosarVer = autosar.SchemaVersionHandler.instance().getSchemaVersion();
           
           switch arDataAccessMode
             case {'ImplicitReceive', 'ExplicitReceive', ...
                   'ExplicitReceiveByVal', 'QueuedExplicitReceive', ...
                   'EndToEndRead'}
               % DaVinci does not allow implicit and explicit access to the same variable in
               % the same runnable (this would be considered good practice
               % although the AUTOSAR standard does allow this). We achieve the
               % same by setting the mode string to be 'Read', rather than
               % Read|IRead|Receive.
               if ~AR4.isCompatibleSchemaVer(autosarVer)
                   switch arDataAccessMode
                       case 'ExplicitReceiveByVal'
                           DAStudio.error('autosarstandard:validation:explicitReceiveByValAutosarVersion', port.Block);
                   end
               end               
               
               if str2double(autosarVer) < 3.2
                   switch arDataAccessMode
                       case 'EndToEndRead'
                           DAStudio.error('autosarstandard:validation:endToEndAutosarVersion', port.Block);
                   end
               end               
               
               apiModeStr = 'Read';
             case {'ImplicitSend', 'ImplicitSendByRef', 'ExplicitSend', 'EndToEndWrite'}
               % DaVinci does not allow implicit and explicit access to same variable to the
               % same runnable (this would be considered good practice although the
               % AUTOSAR standard does allow this). We achieve the same by setting
               % the mode string to be 'Write', rather than Write|IWrite.
               if strcmp(arDataAccessMode, 'EndToEndWrite') && ...
                       str2double(autosarVer) < 3.2
                   DAStudio.error('autosarstandard:validation:endToEndAutosarVersion', port.Block);
               end
               apiModeStr = 'Write';
             case {'ErrorStatus'}
                 % no API for error status
                 return
             case {'IsUpdated'}
                 % no API for is updated
                 return
             case {'ModeReceive', 'ModeSend'}
                apiModeStr = 'Mode';
             case {'BasicSoftwarePort'}
                pName = get_param(port.Block, 'Name');
                DAStudio.error('autosarstandard:validation:BasicSoftwarePortNotSupported', ...
                               pName);
             otherwise
               assert(false, 'Did not recognize data access mode %s', arDataAccessMode);
           end

           apiStr = sprintf('Rte_%s_%s_%s', apiModeStr, arPortName, arElementName);

           slPortName    = get_param(port.Block, 'Name');
           usrReadableRefStr = DAStudio.message('RTW:autosar:simulinkPort', slPortName);
           if apiMap.isKey(apiStr)
               % Found now error              
               dataObj = autosar.api.getAUTOSARProperties(hModel, true);
               componentQName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
               interfaceQName = dataObj.get([componentQName, '/', arPortName], 'Interface', 'PathType', 'FullyQualified');
               
               msg = DAStudio.message('RTW:autosar:duplicateApi',...
                                      usrReadableRefStr, apiMap(apiStr).usrReadableRefStr, arElementName, interfaceQName, arPortName);
               DAStudio.error('RTW:fcnClass:finish', msg);
           else
               % not found: add
               apiMap( apiStr ) = struct( 'usrReadableRefStr', usrReadableRefStr  );
           end
       end
       
       
       function arName = convertToAutosarPlatformTypeName(slTypeName)
       % convert MATLAB builtin type name to Autosar platform type name, passing through any non built-ins
            
           persistent typeMap
            
           if isempty(typeMap)
            
               arNames = {'Boolean', 'Boolean', 'SInt8', 'SInt16', 'SInt32',...
                          'UInt8', 'UInt16', 'UInt32', 'Float', 'Double'};
          
               % MATLAB types
               matlabNames = {'logical', 'boolean', 'int8', 'int16', 'int32',...
                              'uint8', 'uint16', 'uint32', 'single', 'double'};

               typeMap = containers.Map(matlabNames, arNames);
           end
            
           if  typeMap.isKey(slTypeName)
               arName = typeMap(slTypeName);
           else
               arName = slTypeName;
           end
           
       
       end
       
    end
    
end

% LocalWords:  Thru Da Vinci IRead IWrite usr Compu Func SInt UInt blocktype Nv
% LocalWords:  autosarstandard RTB irv noncomplex Nvm Rw Dont SVCE
