function [status, msg, irvLines] = p_checkIrvs(wrapperH, ssBlkH, stopOnError)
% P_CHECKIRVS check the inter-runnable variables contained in ssBlkH
%
%   [STATUS, MSG, IRVLINES] = P_CHECKIRVS(SSBLKH, STOPONERROR)
%       WRAPPERH: handle to wrapper subsystem
%         SSBLKH: subsystem handle
%    STOPONERROR: 1 (ie: when invoked from runValidation) or 0 if called
%                 for extracting all Irv candidates (ie: when invoked from
%                 getDialogSchema or syncWithModel)
%         STATUS: O (error) or 1 
%            MSG: the error message
%       IRVLINES: structure in which each field name maps to a valid
%                 line name and the field's content is the line handle
%

% Copyright 2010-2015 The MathWorks, Inc.
%   

if nargin < 2
    stopOnError = true;
end

% Default outputs
msg = '';
status = 1;
irvLines = containers.Map();

ssBlkH = get_param(ssBlkH, 'Handle');
isSSExpFcn = autosar.validation.ExportFcnValidator.isVirtualSubsysExportFcn(ssBlkH);

% Recurse into virtual subsystems
blocks = find_system(ssBlkH, 'FindAll', 'on', 'SearchDepth', 1, 'FollowLinks', 'on', ...
    'LookUnderMasks', 'all', 'Type', 'block', 'BlockType', 'SubSystem');
for idx = 1:length(blocks)
    if (blocks(idx) ~= ssBlkH) && (strcmpi(get_param(blocks(idx), 'Virtual'), 'on'))
        [status, msg, ssIrvLines] = autosar.validation.p_checkIrvs(wrapperH, blocks(idx), stopOnError);
        if (status ~= 1) && stopOnError
            return
        end
        mapkeys = keys(ssIrvLines);
        for mapidx = 1:length(mapkeys)
            irvLines(mapkeys{mapidx}) = ssIrvLines(mapkeys{mapidx});
        end
        clear ssIrvLines
    end
end

% Check on the lines at this level ...
lines = find_system(ssBlkH, 'FindAll', 'on', 'SearchDepth', 1, 'FollowLinks', 'on', ...
    'LookUnderMasks', 'all', 'Type', 'line');
labels = containers.Map();
for idx = 1:length(lines)
    try
        lineObj = get_param(lines(idx), 'Object');
        srcBlock = lineObj.SrcBlockHandle;
        [dstBlocks, srcLabels, srcLabelBlocks] = ...
            autosar.validation.walkThruVirtualBlockToDsts(lineObj, wrapperH, false);
        
        % Check that line has a source block
        if srcBlock < 0
            if ~isempty(dstBlocks) && dstBlocks(1) > 0
                status = 0;
                if stopOnError
                    msg = DAStudio.message('RTW:autosar:MultiRunnableLineWithNoSrc', ...
                        get_param(lineObj.DstPortHandle, 'PortNumber'), ...
                        RTW.AutosarInterface.p_getTreatedFullName(dstBlocks(1)));
                    return
                else
                    % Skip this line if it has no source or destination
                    continue
                end
            end
        end
        
        % Now, if this line has a virtual source, it will be inspected elsewhere, at its true
        % non-virtual source.  Skip it.  
        if strcmpi(get_param(srcBlock, 'BlockType'), 'From') || ...
                (strcmpi(get_param(srcBlock, 'BlockType'), 'Inport') && ...
                 ~strcmpi(get_param(get_param(srcBlock, 'Parent'), 'Type'), 'block_diagram') && ...
                 strcmpi(get_param(get_param(srcBlock, 'Parent'), 'Virtual'), 'on')) && ...
                (strcmpi(get_param(srcBlock, 'BlockType'), 'Subsystem') && ...
                 strcmpi(get_param(srcBlock, 'Virtual'), 'on'))
            continue
        end
        
        % since this method 'checkIrvs' is called only for code generation,
        % we always guarantee that it is either a short-wire ( for SISO )
        % or replaced with variant merge block for multi input variant
        % source. skip this block since it is virtual
        if (strcmpi(get_param(srcBlock,'BlockType'),'VariantSource') || strcmpi(get_param(srcBlock,'BlockType'),'VariantSink'))
            continue;
        end
        % Check that line has a destination block
        if isempty(dstBlocks) || dstBlocks(1) < 0
            status = 0;
            if stopOnError
                msg = DAStudio.message('RTW:autosar:MultiRunnableLineWithNoDst', ...
                    get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                    RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                return
            else
                % Skip this line
                continue
            end
        end
        
        % If the source is an inport of the wrapper subsystem,
        % Its destinations can only be function-call subsystems or viewing device
        if (strcmp(get_param(srcBlock, 'BlockType'), 'Inport') || ...
                strcmp(get_param(srcBlock, 'BlockType'), 'InportShadow')) && ...
                get_param(get_param(srcBlock, 'Parent'), 'Handle') == wrapperH
            for dstIdx = 1:length(dstBlocks)
                if ~(strcmp(RTW.AutosarInterface.p_checkTrigSSTypeUsingTrigPortBlock(dstBlocks(dstIdx)),'function-call') || ...
                        i_isViewingDevice(dstBlocks(dstIdx)) || ...
                        strcmp(get_param(dstBlocks(dstIdx), 'BlockType'), 'ExportFunctionSpecification') || ...
                        autosar.validation.ExportFcnValidator.isServerSubSys(dstBlocks(dstIdx)) || ...
                        autosar.validation.ExportFcnValidator.isModelWideEvent(dstBlocks(dstIdx)))
                    status = 0;
                    if stopOnError
                        msg = DAStudio.message('RTW:autosar:MultiRunnableNoPassThru', ...
                            RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                        return
                    else
                        % Skip this line
                        continue
                    end
                end
            end
            
            % Else if the source is a merge block
            % It must either be connected (potentially through one or more
            % goto blocks) to a unique outport or
            % 1 or more function-call subsystems or viewing device
            % It cannot select""Allow unequal port widths"
        elseif strcmp(get_param(srcBlock, 'BlockType'), 'Merge')
            connectedToOutport = 0;
            connectedToFCSS = 0;
            for dstIdx = 1:length(dstBlocks)
                if (strcmp(RTW.AutosarInterface.p_checkTrigSSTypeUsingTrigPortBlock(dstBlocks(dstIdx)),'function-call'))
                    connectedToFCSS = connectedToFCSS + 1;
                elseif (strcmp(get_param(dstBlocks(dstIdx), 'BlockType'), 'Outport'))
                    connectedToOutport = connectedToOutport + 1;
                elseif i_isViewingDevice(dstBlocks(dstIdx))
                    % okay
                else
                    status = 0;
                    if stopOnError
                        msg = DAStudio.message('RTW:autosar:MultiRunnableMergeBadDst', ...
                            RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                        return
                    else
                        % Skip this line
                        continue
                    end
                end
            end
            
            % Error if connected to both outports and fcss's
            if (connectedToFCSS > 0) && (connectedToOutport > 0)
                status = 0;
                if stopOnError
                    msg = DAStudio.message('RTW:autosar:MultiRunnableMergeBadDst2', ...
                        RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                    return
                else
                    % Skip this line
                    continue
                end
            end
            
            % Error if connected to more than one outport
            if connectedToOutport > 1
                status = 0;
                if stopOnError
                    msg = DAStudio.message('RTW:autosar:MultiRunnableMergeBadDst3', ...
                        RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                    return
                else
                    % Skip this line
                    continue
                end
            end
            
            % If connected to function call subsystems, is an interrunnable
            % variable, and the label must meet constraints.  Note that we
            % don't worry about connected to merge because the merge block will
            % validate its own label (and make sure any inputs are the same as
            % its output
            if connectedToFCSS > 0
                if isempty(srcLabels)
                    status = 0;
                    if stopOnError
                        msg = DAStudio.message('RTW:autosar:MultiRunnableNoLabel', ...
                            1, RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                        return
                    else
                        % Skip this line
                        continue
                    end
                end
                [uniqueLabels, uniqueLabelIndices] = unique(srcLabels);
                if length(uniqueLabels) ~= 1
                    status = 0;
                    if stopOnError
                        msg = DAStudio.message(...
                            'RTW:autosar:MultiRunnableAmbiguousLabel', ...
                            RTW.AutosarInterface.p_getTreatedFullName(srcLabelBlocks{uniqueLabelIndices(1)}), ...
                            RTW.AutosarInterface.p_getTreatedFullName(srcLabelBlocks{uniqueLabelIndices(2)}));
                        return;
                    else
                        % Skip this line
                        continue
                    end
                end

                % Check Simulink labels are unique
                if (labels.isKey(uniqueLabels{1})==true) && (labels(uniqueLabels{1}) ~= srcBlock)
                     status = 0;
                     if stopOnError
                         msg = DAStudio.message('RTW:autosar:MultiRunnableReusedLabel', ...
                                                 get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                                                 RTW.AutosarInterface.p_getTreatedFullName(srcBlock), ...
                                                 uniqueLabels{1});                         
                         return
                     else
                         % Skip this line because cannot map many lines with same label
                         continue
                     end
                end
                labels(uniqueLabels{1}) = srcBlock;
                irvLines(uniqueLabels{1}) = lineObj;                
            end
            
            % No usage of unequal port widths
            if strcmp(get_param(srcBlock,'AllowUnequalInputPortWidths'), 'on')
                status = 0;
                if stopOnError
                    msg = DAStudio.message('RTW:autosar:MultiRunnableMergeBadWidths', ...
                        RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                    return
                else
                    % Skip this line
                    continue
                end
            end
            
            nInputs = eval(get_param(srcBlock, 'Inputs'));
            srcsOfMerge = zeros(nInputs,1);
            srcBlockLines = get_param(srcBlock, 'LineHandles');
            inputLines = srcBlockLines.Inport;
            for pIdx = 1:nInputs
                line = inputLines(pIdx);
                if line < 0
                    status = 0;
                    if stopOnError
                        msg = DAStudio.message('RTW:autosar:MultiRunnableUnconnectedPort', ...
                            'Inport', ...
                            pIdx,...
                            RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                        return
                    else
                        % Skip this line
                        continue
                    end
                end
                
                [srcsOfMerge(pIdx), candidateLabel] = i_getSrcOfMerge(line, wrapperH);
                % If an interrunnable variable, all inputs and the output
                % must be labeled, and labeled the same
                if connectedToFCSS > 0
                    if isempty(candidateLabel)
                        status = 0;
                        if stopOnError
                            msg = DAStudio.message(...
                                'RTW:autosar:MultiRunnableNoLabel', ...
                                get_param(get_param(line,'SrcPortHandle'),'PortNumber'), ...
                                RTW.AutosarInterface.p_getTreatedFullName(srcsOfMerge(pIdx)));
                            return
                        else
                            % Skip this line
                            continue
                        end
                    end
                    
                    if ~strcmp(uniqueLabels{1}, candidateLabel)
                        status = 0;
                        if stopOnError
                            msg = DAStudio.message(...
                                'RTW:autosar:MultiRunnableInconsistentMergeLabel', ...
                                get_param(get_param(line,'SrcPortHandle'),'PortNumber'), ...
                                RTW.AutosarInterface.p_getTreatedFullName(srcsOfMerge(pIdx)), ...
                                uniqueLabels{1}, ...
                                RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                            return
                        else
                            % Skip this line
                            continue
                        end
                    end
                end
            end
            
            % Each input of the merge must have a unique fcss source
            if ~isequal(size(unique(srcsOfMerge)), size(srcsOfMerge))
                status = 0;
                if stopOnError
                    msg = DAStudio.message('RTW:autosar:MultiRunnableMergeBadSrc', ...
                        RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                    return
                else
                    % Skip this line
                    continue
                end
                
            end
            
            % Else if the source is a function-call subsystem
            % It must either be connected to a unique outport, unique merge, or
            % 1 or more function-call subsystems or viewing device or
            % xxx Todo: Inline variant block (eg: variant source)
        elseif strcmp(RTW.AutosarInterface.p_checkTrigSSTypeUsingTrigPortBlock(srcBlock), ...
                'function-call')
            connectedToMerge = 0;
            connectedToOutport = 0;
            connectedToFCSS = 0;
            connectedToViewingDevice = 0;
            for dstIdx = 1:length(dstBlocks)
                if strcmp(RTW.AutosarInterface.p_checkTrigSSTypeUsingTrigPortBlock(dstBlocks(dstIdx)),...
                        'function-call') || ...
                        autosar.validation.ExportFcnValidator.isServerSubSys(dstBlocks(dstIdx))
                    connectedToFCSS = connectedToFCSS + 1;
                    
                    % If this is a line connecting two function-call
                    % subsystems, it must have a label that is a valid
                    % identifier and is unique for this graph, and it must
                    % be a non-complex scalar
                    if isempty(srcLabels) 
                        if isSSExpFcn && ... 
                                (srcBlock == dstBlocks(dstIdx))
                            % Skip this line
                            continue
                        end
                        status = 0;
                        if stopOnError
                            msg = DAStudio.message('RTW:autosar:MultiRunnableNoLabel', ...
                                get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                                RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                            return
                        else
                            % Skip this line
                            continue
                        end
                    end
                    [uniqueLabels, uniqueLabelIndices] = unique(srcLabels);
                    if length(uniqueLabels) ~= 1
                        status = 0;
                        if stopOnError
                            msg = DAStudio.message(...
                                'RTW:autosar:MultiRunnableAmbiguousLabel', ...
                                RTW.AutosarInterface.p_getTreatedFullName(srcLabelBlocks{uniqueLabelIndices(1)}), ...
                                RTW.AutosarInterface.p_getTreatedFullName(srcLabelBlocks{uniqueLabelIndices(2)}));
                            return;
                        else
                            % Skip this line
                            continue
                        end
                    end
                                       
                    % Check Simulink labels are unique
                    if (labels.isKey(uniqueLabels{1})==true) && (labels(uniqueLabels{1}) ~= srcBlock)
                        status = 0;
                        if stopOnError
                            msg = DAStudio.message('RTW:autosar:MultiRunnableReusedLabel', ...
                                                    get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                                                    RTW.AutosarInterface.p_getTreatedFullName(srcBlock), ...
                                                    uniqueLabels{1});                         
                            return
                        else
                            % Skip this line because cannot map many lines with same label
                            continue
                        end
                   end
                   labels(uniqueLabels{1}) = srcBlock;
                   irvLines(uniqueLabels{1}) = lineObj;                    
                    
                elseif strcmp(get_param(dstBlocks(dstIdx), 'BlockType'), ...
                        'Outport')
                    connectedToOutport = connectedToOutport + 1;
                elseif strcmp(get_param(dstBlocks(dstIdx), 'BlockType'), ...
                        'Merge')
                    connectedToMerge = connectedToMerge + 1;
                elseif i_isViewingDevice(dstBlocks(dstIdx))
                    connectedToViewingDevice = connectedToViewingDevice + 1;
                else
                    status = 0;
                    if stopOnError
                        msg = DAStudio.message('RTW:autosar:MultiRunnableBadDst', ...
                            get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                            RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                        return
                    else
                        % Skip this line
                        continue
                    end
                    
                end
            end
            
            % Error if connected to both merges and FCSS
            if  connectedToFCSS > 0 && connectedToMerge > 0
                status = 0;
                if stopOnError
                    msg = DAStudio.message('RTW:autosar:MultiRunnableBadDst2', ...
                        get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                        RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                    return
                else
                    % Skip this line
                    continue
                end
            end
            
            % Error if connected to root outport and SS when in right click
            % build
            if isSSExpFcn && connectedToFCSS > 0 && connectedToOutport > 0
                status = 0;
                if stopOnError
                    msg = DAStudio.message('RTW:autosar:MultiRunnableBadDst7', ...
                        get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                        RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                    return
                else
                    % Skip this line
                    continue
                end
            end
            
            % Error if connected to more than one outport
            if connectedToOutport > 1
                status = 0;
                if stopOnError
                    msg = DAStudio.message('RTW:autosar:MultiRunnableBadDst3', ...
                        get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                        RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                    return
                else
                    % Skip this line
                    continue
                end
            end
            
            % Error if connected to more than one merge
            if connectedToMerge > 1
                status = 0;
                if stopOnError
                    msg = DAStudio.message('RTW:autosar:MultiRunnableBadDst4', ...
                        get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                        RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                    return
                else
                    % Skip this line
                    continue
                end
            end
            
            % Error if connected to both merge and outports
            if connectedToOutport > 0 && connectedToMerge > 0
                status = 0;
                if stopOnError
                    msg = DAStudio.message('RTW:autosar:MultiRunnableBadDst5', ...
                        get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                        RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                    return
                else
                    % Skip this line
                    continue
                end
                
            end
            
            if connectedToViewingDevice > 0 && connectedToFCSS == 0 && ...
                    connectedToOutport == 0 && connectedToMerge == 0 && ...
                    strcmp(lineObj.SegmentType, 'trunk')
                status = 0;
                if stopOnError
                    msg = DAStudio.message('RTW:autosar:MultiRunnableBadDst6', ...
                        get_param(lineObj.SrcPortHandle, 'PortNumber'), ...
                        RTW.AutosarInterface.p_getTreatedFullName(srcBlock));
                    return
                else
                    % Skip this line
                    continue
                end
                
            end
            
        end
        
    catch me
        % Got an unexpected error while extracting the line information
        status = 0;
        if stopOnError
            msg = me.message;
            return
        else
            % Skip this line
            continue
        end
    end
end


%--------------------------------------------------------------------------
function isValid = i_isViewingDevice( blk )
% is the block a viewing device?

switch get_param(blk, 'BlockType')
    case {'Scope', 'Display'}
        isValid = true;
    otherwise
        isValid = false;
end

function [srcOfMerge, srcLabel] = i_getSrcOfMerge(line, wrapperH)

lineObj = get_param(line, 'Object');
srcBlock = lineObj.SrcBlockHandle;
srcLabel = '';
if (srcBlock > 0 && ...
    strcmp(get_param(srcBlock, 'BlockType'), 'From'))
    blkObj = get_param(srcBlock, 'Object');
    gotoHandle = blkObj.GotoBlock.handle;
    gotoPortHandles = get_param(gotoHandle, 'PortHandles');
    gotoInPortHandle = gotoPortHandles.Inport;
    gotoLineHandle = get_param(gotoInPortHandle, 'Line');
    [srcOfMerge, candidateLabel] =  i_getSrcOfMerge(gotoLineHandle, wrapperH);
    if ~isempty(candidateLabel)
        srcLabel = candidateLabel;
    end
elseif (srcBlock > 0 && ...
        strcmpi(get_param(srcBlock, 'BlockType'), 'Subsystem') && ...
        strcmp(get_param(srcBlock, 'virtual'), 'on'))
    outPortHdl = lineObj.SrcPortHandle;
    outPortIdx = num2str(get_param(outPortHdl, 'PortNumber'));
    outPort = find_system(srcBlock, 'SearchDepth', 1, 'LookUnderMasks', 'all', ...
                          'FollowLinks', 'on', 'Type', 'Block', ...
                          'BlockType','Outport','Port', outPortIdx);
    outPortPortHandles = get_param(outPort, 'PortHandles');
    outPortOutPortHandle = outPortPortHandles.Inport(1);
    outPortLineHandle = get_param(outPortOutPortHandle, 'Line');
    [srcOfMerge, candidateLabel] = i_getSrcOfMerge(outPortLineHandle, wrapperH);
    if ~isempty(candidateLabel)
        srcLabel = candidateLabel;
    end
elseif (srcBlock > 0 && ...
        strcmp(get_param(srcBlock, 'BlockType'), 'Inport') && ...
        strcmp(get_param(get_param(srcBlock, 'Parent'), 'Type'), 'block')) && ...
        get_param(get_param(srcBlock, 'Parent'), 'Handle') ~= wrapperH
    ssParent = get_param(srcBlock, 'Parent');
    ssParentPortIdx = str2double(get_param(srcBlock, 'Port'));
    ssParentPortHandles = get_param(ssParent, 'PortHandles');
    ssParentInPortHandle = ssParentPortHandles.Inport(ssParentPortIdx);
    ssParentLineHandle = get_param(ssParentInPortHandle, 'Line');
    [srcOfMerge, candidateLabel] =  i_getSrcOfMerge(ssParentLineHandle, wrapperH);
    if ~isempty(candidateLabel)
        srcLabel = candidateLabel;
    end
else
    srcOfMerge = srcBlock;
end
% Accumulate any existing srcLabel info regardless of the destination block type
if ~isempty(lineObj.Name)
    srcLabel = lineObj.Name;
end

% LocalWords:  Autosar autosar Thru fcss's interrunnable fcss one's Froms Dsts
% LocalWords:  IRVLINES SSBLKH STOPONERROR ie WRAPPERH thru eg

