function [modelH, success] = createComponentAsModel(this, ComponentName, varargin)
%CREATECOMPONENTASMODEL creates an AUTOSAR Atomic-Software-Component as a 
%   Simulink model.
%
%   createComponentAsModel(IMPORTEROBJ, 'COMPONENTNAME') creates the Simulink
%   model corresponding to the AUTOSAR Atomic-Software-Component 'COMPONENT' 
%   described in the XML file imported by the arxml.importer object IMPORTEROBJ.
%
%   Usage:
%      [modelH, success] = createComponentAsModel(importerObj,...
%                                                 componentName, varargin)
%
%   Inputs:
%      importerObj:   arxml.importer object
%      componentName: absolute Short-Name path of the Atomic-Software-Component
%      varargin:      Parameter/Value pairs. Supported parameters are:
%
%         AutoSave: true or false (default). If it is true, this function
%         automatically saves the generated Simulink model.
%
%         CreateInternalBehavior: true or false (default). By default, the
%         function imports a multi-runnable AUTOSAR software component and maps
%         it into a new rate-based model. If the arxml code contains periodic
%         runnables, the importer adds rate-based model content, including
%         periodic functions and data transfer lines with rate transitions, and
%         maps them to corresponding periodic runnables and interrunnable
%         variables (IRVs) imported from the AUTOSAR software component. If
%         true, the function imports a multi-runnable AUTOSAR software component
%         and maps it into a new function-call-based model. The importer adds
%         function-call subsystem or function blocks and signal lines in the
%         model and maps them to corresponding runnables and IRVs imported from
%         the AUTOSAR software component.
%
%         CreateSimulinkObject: true (default) or false. If it is true, this
%         function creates the Simulink.AliasType and Simulink.NumericType
%         corresponding to the AUTOSAR data types found in the XML file.
%
%         DataDictionary: String specifying a Simulink data dictionary into
%         which to import data objects corresponding to AUTOSAR data types in
%         the XML file. The model becomes associated with the specified data
%         dictionary.
% 
%         InitializationRunnable: String specifying the name of an existing
%         runnable as the initialization runnable for the component.
%
%         NameConflictAction: 'overwrite' (default) or 'makenameunique' or
%         'error'. Use this parameter to control the action to take when a
%         Simulink model with the same name as the component already exists.
%
%   Outputs:
%      modelH:  Model handle.
%      success: It is true if this function is successful. Otherwise, 
%               it is false.
%
%   See also arxml.importer.getComponentNames
%      arxml.importer.getFile arxml.importer.setFile 
%      arxml.importer.getDependencies arxml.importer.setDependencies

%   Copyright 2007-2015 The MathWorks, Inc.

  
success = false; %#ok
modelH  = -1;    %#ok
  
% Create an instance of the inputParser class.
argParser = inputParser;   

% Add arguments to the schema
argParser.addRequired('this', @(x) isa(x, class(x)));
argParser.addRequired('ComponentName', @ischar);
argParser.addParameter('CreateSimulinkObject', true, @(x)(islogical(x)||(x==1)||(x==0)));
argParser.addParameter('NameConflictAction', 'overwrite', @(x)any(strcmpi(x,{'overwrite','makenameunique','error'})));
argParser.addParameter('AutoSave', false, @(x)(islogical(x)||(x==1)||(x==0)));
argParser.addParameter('CreateInternalBehavior', false, @(x)(islogical(x)||(x==1)||(x==0)));
argParser.addParameter('InitializationRunnable', '', @(x)(ischar(x)));
argParser.addParameter('DataDictionary', '', @(x)(ischar(x)));
argParser.addParameter('PredefinedVariant', '', @(x)(ischar(x)));
argParser.addParameter('SystemConstValueSets', '', @(x)(iscell(x)));

% Read and validate each argument
argParser.parse(this, ComponentName, varargin{:});

% Check if the DataDictionary exists
autosar.mm.mm2sl.utils.checkAndCreateDD(argParser.Results.DataDictionary)
% Create the component
[modelH, success] = p_create_slcomponent(this, argParser, true);

% LocalWords:  IMPORTEROBJ COMPONENTNAME makenameunique BSP
